from .Component import Component
from .Sample import Sample


class Container(Component):
    """
    Entity class holding state of any hierarchical sample container
    """

    def __init__(self, type, container, address, scannable):
        super(Container, self).__init__(container, address, scannable)
        self.type = type
        self.components = []

    #########################           PUBLIC           #########################

    def get_type(self):
        """
        Returns a description of the type of container
        Known types:
        -    Puck
        -    Vial
        -    Plate
        -    Well
        -    SC3
        -    PlateSupport
        -    GRob
        :rtype: str
        """
        return self.type

    def get_components(self):
        """
        Returns the list of components of this container
        :rtype: list
        """
        return self.components

    def get_number_of_components(self):
        return len(self.components)

    def get_sample_list(self):
        """
        Returns the list of all Sample objects under of this container (recursively)
        :rtype: list
        """
        samples = []
        for c in self.get_components():
            if isinstance(c, Sample):
                samples.append(c)
            else:
                samples.extend(c.get_sample_list())
        return samples

    def get_basket_list(self):
        basket_list = []
        for basket in self.components:
            if isinstance(basket, Basket):
                basket_list.append(basket)
        return basket_list

    def get_present_samples(self):
        """
        Returns the list of all Sample objects under of this container (recursively) tagged as present
        :rtype: list
        """
        ret = []
        for sample in self.get_sample_list():
            if sample.is_present():
                ret.append(sample)

    def is_empty(self):
        """
        Returns true if there is no sample present sample under this container
        :rtype: bool
        """
        for s in self.get_sample_list():
            if s.is_present():
                return False
        return True

    def get_component_by_address(self, address):
        """
        Returns a component through its slot address or None if address is invalid
        :rtype: Component
        """
        for c in self.get_components():
            if c.get_address() == address:
                return c
            if isinstance(c, Container):
                aux = c.get_component_by_address(address)
                if aux is not None:
                    return aux
        return None

    def has_component_address(self, address):
        """
        Returns if has a component with a given address
        :rtype: bool
        """
        return self.get_component_by_address(address) is not None

    def get_component_by_id(self, id):
        """
        Returns a component through its id or None if id is invalid
        :rtype: Component
        """
        for c in self.get_components():
            if c.get_id() == id:
                return c
            if isinstance(c, Container):
                aux = c.get_component_by_id(id)
                if aux is not None:
                    return aux
        return None

    def has_component_id(self, id):
        """
        Returns if has a component with a given ID
        :rtype: bool
        """
        return self.get_component_by_id(id) is not None

    def get_selected_sample(self):
        for s in self.get_sample_list():
            if s.is_selected():
                return s
        return None

    def get_selected_component(self):
        for c in self.get_components():
            if c.is_selected():
                return c
        return None

    def clear_info(self):
        Component._reset_dirty(self)
        for c in self.get_components():
            c.clear_info()

    #########################           PROTECTED           #########################

    def _add_component(self, c):
        self.components.append(c)

    def _remove_component(self, c):
        self.components.remove(c)

    def _clear_components(self):
        self.components = []

    def _reset_dirty(self):
        Component._reset_dirty(self)
        for c in self.get_components():
            c._reset_dirty()

    def _set_selected_sample(self, sample):
        for s in self.get_sample_list():
            if s == sample:
                s._set_selected(True)
            else:
                s._set_selected(False)

    def _set_selected_component(self, component):
        if component is None:
            for c in self.get_components():
                c._set_selected(False)
        else:
            component._set_selected(True)

    def _set_selected(self, selected):
        if not selected:
            for c in self.get_components():
                c._set_selected(False)
        Component._set_selected(self, selected)


class Basket(Container):
    __TYPE__ = "Puck"

    def __init__(self, container, number, samples_num=10, name="Puck"):
        super(Basket, self).__init__(
            self.__TYPE__, container, Basket.get_basket_address(number), True
        )

        self._name = name
        self.samples_num = samples_num
        for i in range(samples_num):
            slot = Pin(self, number, i + 1)
            self._add_component(slot)

    @staticmethod
    def get_basket_address(basket_number):
        return str(basket_number)

    def get_number_of_samples(self):
        return self.samples_num

    def clear_info(self):
        # self.get_container()._reset_basket_info(self.get_index()+1)
        self.get_container()._trigger_info_changed_event()


class Pin(Sample):
    STD_HOLDERLENGTH = 22.0

    def __init__(self, basket, basket_no, sample_no):
        super(Pin, self).__init__(
            basket, Pin.get_sample_address(basket_no, sample_no), False
        )
        self._set_holder_length(Pin.STD_HOLDERLENGTH)

    def get_basket_no(self):
        return self.get_container().get_index() + 1

    def get_vial_no(self):
        return self.get_index() + 1

    @staticmethod
    def get_sample_address(basket_number, sample_number):
        return "%s:%02d" % (basket_number, sample_number)



#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

from typing import List

from mxcubecore.HardwareObjects import edna_test_data
from mxcubecore.HardwareObjects.EDNACharacterisation import EDNACharacterisation
from mxcubecore.HardwareObjects.XSDataMXCuBEv1_4 import (
    XSDataInputMXCuBE,
    XSDataResultMXCuBE,
)
from mxcubecore.model import queue_model_objects as qmo

__credits__ = ["MXCuBE collaboration"]
__license__ = "LGPLv3"


class EDNACharacterisationMockup(EDNACharacterisation):
    def __init__(self, name) -> None:
        super(EDNACharacterisationMockup, self).__init__(name)

    def input_from_params(self, data_collection, char_params) -> XSDataInputMXCuBE:
        return XSDataInputMXCuBE.parseString(self.edna_default_input)

    def characterise(self, edna_input) -> XSDataResultMXCuBE:
        return XSDataResultMXCuBE.parseString(edna_test_data.EDNA_RESULT_DATA)

    def is_running(self) -> bool:
        return False

    def dc_from_output(
        self, edna_result, reference_image_collection
    ) -> List[qmo.DataCollection]:
        return []

    def get_default_characterisation_parameters(self) -> qmo.CharacterisationParameters:
        return super(
            EDNACharacterisationMockup, self
        ).get_default_characterisation_parameters()



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.
"""
ID30-B Beam Definer.

Example xml configuration:

.. code-block:: xml

 <object class="ESRF.ID30BBeamDefiner">
   <object href="/udiff_aperture" role="controller"/>
   <object href="/bliss" role="bliss"/>
   <transfocator>tf</transfocator>
   <beam_config>
      <name>A10</name>
      <beam_size>0.01, 0.05</beam_size>
      <aperture_size>10</aperture_size>
      <defocused_beam>False</defocused_beam>
   </beam_config>
   <beam_config>
      <name>A20 um</name>
      <beam_size>0.02, 0.02</beam_size>
      <aperture_size>20</aperture_size>
      <defocused_beam>False</defocused_beam>
   </beam_config>
   <beam_config>
      <name>A30</name>
      <beam_size>0.03, 0.03</beam_size>
      <aperture_size>30</aperture_size>
      <defocused_beam>False</defocused_beam>
   </beam_config>
   <beam_config>
      <name>A50</name>
      <beam_size>0.05, 0.05</beam_size>
      <aperture_size>50</aperture_size>
      <defocused_beam>True</defocused_beam>
   </beam_config>
   <beam_config>
      <name>A75</name>
      <beam_size>0.075, 0.075</beam_size>
      <aperture_size>75</aperture_size>
      <defocused_beam>True</defocused_beam>
   </beam_config>
 </object>
"""

__copyright__ = """ Copyright © by the MXCuBE collaboration """
__license__ = "LGPLv3+"


from enum import Enum

from mxcubecore.HardwareObjects.ESRF.ESRFBeamDefiner import ESRFBeamDefiner


class ID30BBeamDefiner(ESRFBeamDefiner):
    """ID30B beam definer implementation"""

    def __init__(self, *args):
        super().__init__(*args)
        self.controller = None
        self.defocused_beam = None
        self.tf_obj = None

    def init(self):
        """Initialisation"""
        super().init()

        self.controller = self.get_object_by_role("controller")
        bliss = self.get_object_by_role("bliss")
        self.tf_obj = getattr(bliss, self.get_property("transfocator"))
        _dict = {}

        for beam_cfg in self.bd_config:
            name = beam_cfg.get_property("name")
            _ap_size = beam_cfg.get_property("aperture_size")
            _aperture = self.controller.value_to_enum(_ap_size, idx=1)
            _defocused_beam = bool(beam_cfg.get_property("defocused_beam"))
            _dict[name] = {"aperture": _aperture, "defocus": _defocused_beam}

        self.beam_config = _dict
        self._initialise_values()
        self.connect(self.controller, "valueChanged", self._update_name)

        self.defocused_beam = _dict[self.get_current_position_name()]["defocus"]

    def get_state(self):
        """Get the device state.
        Returns:
            (enum 'HardwareObjectState'): Device state.
        """
        return self.controller.get_state()

    def _update_name(self, value=None):
        """
        Emits:
            valueChanged (str): Current beam size name.
        """
        name = self.get_current_position_name()
        self.emit("valueChanged", name)

    def get_current_position_name(self):
        """Get the current beam size name.
        Returns:
            (str): Current beam size name.
        """
        _aperture = self.controller.get_value()
        for name, value in self.beam_config.items():
            if value["aperture"] == _aperture:
                return name
        return "UNKNOWN"

    def get_value(self):
        """Get the device value
        Returns:
            (Enum): The current position Enum.
        """
        try:
            value = self.VALUES[self.get_current_position_name()]
            if isinstance(value.value, tuple):
                return value
            return Enum("Dummy", {value.name: value.value[0]})[value.name]
        except (ValueError, KeyError):
            return self.VALUES.UNKNOWN

    def set_value(self, value, timeout=None):
        """Set the beam size.
        Args:
            value(str): name of the beam size to set.
            timeout(float): Timeout to wait for the execution to finish [s].
        Raises:
            RuntimeError: Cannot change beam size.
        """
        if isinstance(value, Enum):
            value = value.name

        self.controller.set_value(self.beam_config[value]["aperture"], timeout=timeout)
        self.defocused_beam = self.beam_config[value]["defocus"]
        self.tf_obj.change_tf(defocus=self.defocused_beam)



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE.  If not, see <http://www.gnu.org/licenses/>.

"""
[Name] Harvester

[Description]
Harvester mockup is use as a replacement of the Dewar sample storage
This hardware object is use in couple with a Sample changer .
and it is compatible with the Crystal Direct Harvester 3.
It has some functionalities, like Harvest Sample, .

[Commands]

 - getSampleList : Get list of available sample from Harvester
 - Harvest : load sample from Harvester

-----------------------------------------------------------------------
"""

import logging

import gevent

from mxcubecore.BaseHardwareObjects import HardwareObject


class HarvesterMockup(HardwareObject):
    """Harvester functionality

    The Harvester Class consists of methods executing exporter commands
    to communicate with the Crystal Direct Harvester Machine

    """

    __TYPE__ = "Harvester"

    def __init__(self, name):
        super().__init__(name)
        self.timeout = 3  # default timeout

        # Internal variables -----------
        self.calibration_state = False
        self.room_temperature_mode = False

    def init(self):
        """Init"""
        self.exporter_addr = self.get_property("exporter_address")

    def set_calibration_state(self, state):
        """Set Calibration state

        Args:
        state (bool) : Whether the a calibration was perform
        """

        self.calibration_state = state

    def _wait_ready(self, timeout=None):
        """Wait Harvester to be ready

        Args:
        (timeout) : Whether to wait for a amount of time
        None means wait forever timeout <=0 use default timeout
        """
        if timeout is not None and timeout <= 0:
            timeout = self.timeout

        err_msg = "Timeout waiting for Harvester to be ready"

        with gevent.Timeout(timeout, RuntimeError(err_msg)):
            while not self._ready():
                logging.getLogger("user_level_log").info(
                    "Waiting Harvester to be Ready"
                )
                gevent.sleep(3)

    def _wait_sample_transfer_ready(self, timeout=None):
        """Wait Harvester to be ready to transfer a sample

        Args:
        timeout (second) : Whether to wait for a amount of time
        None means wait forever timeout <=0 use default timeout
        """
        if timeout is not None and timeout <= 0:
            timeout = self.timeout

        err_msg = "Timeout waiting for Harvester to be ready to transfer"

        with gevent.Timeout(timeout, RuntimeError(err_msg)):
            while not self._ready_to_transfer():
                logging.getLogger("user_level_log").info(
                    "Waiting Harvester to be ready to transfer"
                )
                gevent.sleep(3)

    def _execute_cmd_exporter(self, cmd, *args, **kwargs):
        """Exporter Command implementation

        Args:
        cmd (string) : command type
        args, kwargs (string): commands arguments, and  command or attribute

        return : respond
        """
        ret = None
        if args:
            args_str = "%s" % "\t".join(map(str, args))
        if kwargs.pop("command", None):
            exp_cmd = self.add_command(
                {
                    "type": "exporter",
                    "exporter_address": self.exporter_addr,
                    "name": "%s" % cmd,
                },
                "%s" % cmd,
            )
            if args:
                ret = exp_cmd(args_str)
            else:
                ret = exp_cmd()
        if kwargs.pop("attribute", None):
            exp_attr = self.add_channel(
                {
                    "type": "exporter",
                    "exporter_address": self.exporter_addr,
                    "name": "%s" % cmd,
                },
                "%s" % cmd[3:],
            )
            if cmd.startswith("get"):
                return exp_attr.get_value()
            if cmd.startswith("set"):
                ret = exp_attr.set_value(args_str)

        return ret

    # ---------------------- State --------------------------------

    def get_state(self):
        """Get the Harvester State

        Return (str):  state "Ready, Running etc.."
        """
        return "Ready"

    def get_status(self):
        """Get the Harvester Status

        Return (str):  Status
        """
        # return self._execute_cmd_exporter("getStatus", attribute=True)
        return "Ready"

    def _ready(self):
        """check whether the Harvester is READY

        Return (bool):  True if Harvester is Ready otherwise False
        """
        return True

    def _busy(self):
        """check whether the Harvester is BUSY

        Return (bool):  True if Harvester is not Ready otherwise False
        """
        return not self._ready()

    def _ready_to_transfer(self):
        """check whether the Harvester is Waiting Sample Transfer

        Return (bool):  True if Harvester is Waiting Sample Transfer otherwise False
        """
        # return self._execute_cmd_exporter("getStatus", attribute=True) == "Waiting Sample Transfer"
        return "Waiting Sample Transfer"

    def get_samples_state(self):
        """Get the Harvester Samples State

        Return (List):  list of crystal state "waiting_for_transfer, Running etc.."
        """
        # return self._execute_cmd_exporter("getSampleStates", command=True)
        return ["ready_to_execute", "needs_repositionning", "failed", "harvested"]

    def get_current_crystal(self):
        """Get the Harvester current harvested crystal

        Return (str): the crystal uuid
        """
        # return self._execute_cmd_exporter("getCurrentSampleID", attribute=True)
        return None

    def is_crystal_harvested(self, crystal_uuid):
        """Check Whether if the current crystal is harvested

        args: the crystal uuid

        Return (bool):  True if the crystal is the current harvested crystal
        """
        res = False
        in_list = crystal_uuid in self.get_crystal_uuids()
        if in_list:
            Current_SampleID = self.get_current_crystal()
            if crystal_uuid == Current_SampleID:
                res = True
        return res

    def current_crystal_state(self, crystal_uuid):
        """get current crystal state

        Args:
        state (str) : Crystal uuid

        Return (str):  State of the crystal uuid
        """
        sample_states = self.get_samples_state()
        crystal_uuids = self.get_crystal_uuids()

        for index, x_tal in enumerate(crystal_uuids):
            if crystal_uuid == x_tal:
                return sample_states[index]

        return None

    def check_crystal_state(self, crystal_uuid):
        """Check whether if a Crystal is in pending_and_current or not

        Args (str) : Crystal uuid

        Return (str):  status of the crystal_uuid pending / current
        """
        sample_states = self.get_samples_state()
        crystal_uuids = self.get_crystal_uuids()

        for index, x_tal in enumerate(crystal_uuids):
            if crystal_uuid == x_tal and sample_states[index] == "waiting_for_transfer":
                return "pending_and_current"
            elif (
                crystal_uuid != x_tal and sample_states[index] == "waiting_for_transfer"
            ):
                return "pending_not_current"
            else:
                return None

    def get_crystal_uuids(self):
        """Get the Harvester Sample List uuid

        Return (List):  list of crystal by uuid from the current processing plan"
        """
        # harvester_crystal_list = self._execute_cmd_exporter("getSampleList",  attribute=True)
        return [
            "c9ca5e0d-35fd-4ff9-9ad2-4de9fd47fd83",
            "94730c39-bf66-416f-ab97-f755e45f6a3a",
            "5d5dfc88-009d-4486-8526-33bf58c4d5d9",
            "46aca4a4-0c00-4685-b1e1-53071bdcb66d",
        ]

    def get_sample_names(self):
        """Get the Harvester Sample List Name

        Return (List):  list of crystal by names from the current processing plan"
        """
        # harvester_sample_names = self._execute_cmd_exporter("getSampleNames", attribute=True)
        return ["TestSample1", "TestSample2", "TestSample3", "TestSample4"]

    def get_crystal_images_urls(self, crystal_uuid):
        """Get the Harvester Sample List Images

        Args (str) : Crystal uuid

        Return (List):  list of crystal by image_url from current processing plan"
        """
        # crystal_images_url = self._execute_cmd_exporter("getImageURL", crystal_uuid, command=True)
        return "https://htxlab.embl.org//rawimages/2021//CD032401/6/FORMULATRIX_CD032401_6_08-09-2021_05_02_01_00_99_Vis.jpg"

    def get_sample_acronyms(self):
        """Get the Harvester Sample List by Acronyms

        Return (List):  list of crystal by Acronyms from the current processing plan"
        """
        # harvester_sample_acronyms = self._execute_cmd_exporter("getSampleAcronyms", attribute=True)
        return ["cryoprotectant", "cryoprotectant", "cryoprotectant", "cryoprotectant"]

    # ------------------------------------------------------------------------------------

    def abort(self):
        """Send Abort command
        Abort any current Harvester Actions
        """
        # return self._execute_cmd_exporter("abort", command=True)
        return None

    def harvest_crystal(self, crystal_uuid):
        """Harvester crystal

        Args (str) : Crystal uuid
        """
        # return self._execute_cmd_exporter("harvestCrystal", crystal_uuid, command=True)
        return False

    def transfer_sample(self):
        """Transfer the current Harvested Crystal"""
        # return self._execute_cmd_exporter("startTransfer", command=True)
        return False

    def trash_sample(self):
        """Trash the current Harvested Crystal"""
        # return self._execute_cmd_exporter("trashSample", command=True)
        return False

    # -----------------------------------------------------------------------------

    def load_plate(self, plate_id):
        """Load a plate from Harvester to MD

        Args (str) : Plate ID
        """
        # return self._execute_cmd_exporter("loadPlate", plate_id, command=True)
        return

    def get_plate_id(self):
        """get current plate ID

        Args:
        Return (str) : current Plate ID
        """
        # return self._execute_cmd_exporter("getPlateID", attribute=True)
        return ""

    def get_image_target_x(self, crystal_uuid):
        """Get the crystal images position x

        Args (str) : Crystal uuid

        Return (float):  Crystal x coordinate in plate
        """
        # return self._execute_cmd_exporter("getImageTargetX", crystal_uuid, command=True)
        return 26.868617890692

    def get_image_target_y(self, crystal_uuid):
        """Get the crystal images position Y

        Args (str) : Crystal uuid

        Return (float): Crystal Y coordinate in plate
        """
        # return self._execute_cmd_exporter("getImageTargetY", crystal_uuid, command=True)
        return 50.122652377553

    def get_room_temperature_mode(self):
        """get  RoomTemperature Mode state

        Args (str) : Crystal uuid

        Return (bool):  TemperatureMode , True if Room Temp else False
        """
        # return self._execute_cmd_exporter("getRoomTemperatureMode", attribute=True)
        return True

    def set_room_temperature_mode(self, value):
        """Set Harvester temperature mode

        Args: (bool) set room temperature when true
        """
        # self._execute_cmd_exporter("setRoomTemperatureMode", value, command=True)
        print("setting HA Room temperature to: %s" % value)
        return self.get_room_temperature_mode()

    # -------------------- Calibrate  Drift Shape offset ----------------------------

    def get_last_sample_drift_offset_x(self):
        """Sample Offset X position when drifted
        Return (float):  last pin drift offset x
        """
        # last_sample_drift_offset_x = self._execute_cmd_exporter("getLastSampleDriftOffsetX", attribute=True)
        return

    def get_last_sample_drift_offset_y(self):
        """Sample Offset Y position when drifted
        Return (float):  last pin drift offset y
        """
        # last_sample_drift_offset_y = self._execute_cmd_exporter("getLastSampleDriftOffsetY", attribute=True)
        return

    def get_last_sample_drift_offset_z(self):
        """Sample Offset Z position when drifted
        Return (float):  last pin drift offset z
        """
        # pin_last_drift_offset_z = self._execute_cmd_exporter("getLastSampleDriftOffsetZ", attribute=True)
        return

    # ---------------------- Calibrate Cut Shape offset----------------------------

    def get_last_pin_cut_shape_offset_x(self):
        """Pin shape Offset x position when
        Return (float):  last pin cut shape offset x
        """
        # pin_last_cut_shape_offset_x = self._execute_cmd_exporter("getLastSampleCutShapeOffsetX", attribute=True)
        return

    def get_last_pin_cut_shape_offset_y(self):
        """Pin shape Offset Y position when
        Return (float):  last pin cut shape offset y
        """
        # pin_last_cut_shape_offset_y = self._execute_cmd_exporter("getLastSampleCutShapeOffsetY", attribute=True)
        return

    # =============== Pin / Calibration -----------------------------

    def load_calibrated_pin(self):
        """Start Pin Calibration Procedure"""
        # return self._execute_cmd_exporter("loadCalibratedPin", command=True)
        return

    def store_calibrated_pin(self, x, y, z):
        """Store x , y , z offsets position after calibration procedure

        Args: (float) x, y, z offsets
        """
        # return self._execute_cmd_exporter("storePinToBeamOffset", x , y , z,  command=True)
        return

    def get_calibrated_pin_offset(self):
        """Get Stored x , y , z offsets position after calibration procedure

        return: (float) x, y, z offsets
        """
        # pin_to_beam_offset = self._execute_cmd_exporter("getPinToBeamOffset", command=True)
        return

    def get_number_of_available_pin(self):
        """Get number of available pin

        return: (Integer)
        """
        # return self._execute_cmd_exporter("getNbRemainingPins", command=True)
        return 1

    def get_offsets_for_sample_centering(self):
        pin_to_beam = tuple(self.get_calibrated_pin_offset())

        sample_drift_x = float(self.get_last_sample_drift_offset_x())
        sample_drift_y = float(self.get_last_sample_drift_offset_y())
        sample_drift_z = -float(self.get_last_sample_drift_offset_z())

        pin_cut_shape_x = float(self.get_last_pin_cut_shape_offset_x())
        pin_cut_shape_y = float(self.get_last_pin_cut_shape_offset_y())

        phiy_offset = sample_drift_x - pin_cut_shape_x + float(pin_to_beam[1])

        centringFocus = sample_drift_z + float(pin_to_beam[0])

        centringTableVertical = sample_drift_y - pin_cut_shape_y + float(pin_to_beam[2])

        return (phiy_offset, centringFocus, centringTableVertical)



"""
A client for ISPyB Webservices.
"""

import logging
import warnings

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.ProposalTypeISPyBLims import ProposalTypeISPyBLims
from mxcubecore.model.lims_session import (
    LimsSessionManager,
    Session,
)

# to simulate wrong loginID, use anything else than idtest
# to simulate wrong psd, use "wrong" for password
# to simulate ispybDown, but ldap login succeeds, use "ispybDown" for password
# to simulate no session scheduled, use "nosession" for password


LOGIN_TYPE_FALLBACK = "proposal"


class ISPyBClientMockup(ProposalTypeISPyBLims):
    """
    Web-service client for ISPyB.
    """

    def __init__(self, name):
        super().__init__(name)

        self.loginType = LOGIN_TYPE_FALLBACK

    def init(self):
        try:
            self.base_result_url = self.get_property("base_result_url").strip()
        except AttributeError:
            pass

        self.__test_proposal = {
            "status": {"code": "ok"},
            "Person": {
                "personId": 1,
                "laboratoryId": 1,
                "login": None,
                "familyName": "operator on IDTESTeh1",
            },
            "Proposal": {
                "code": "idtest",
                "title": "operator on IDTESTeh1",
                "personId": 1,
                "number": "0",
                "proposalId": 1,
                "type": "MX",
            },
            "Session": [
                {
                    "scheduled": 0,
                    "startDate": "2013-06-11 00:00:00",
                    "endDate": "2023-06-12 07:59:59",
                    "beamlineName": self.beamline_name,
                    "timeStamp": "2013-06-11 09:40:36",
                    "comments": "Session created by the BCM",
                    "sessionId": 34591,
                    "proposalId": 1,
                    "nbShifts": 3,
                }
            ],
            "Laboratory": {"laboratoryId": 1, "name": "TEST eh1"},
        }

        self.loginType = self.get_property("loginType", LOGIN_TYPE_FALLBACK)

    def get_login_type(self):
        warnings.warn(
            "Deprecated method `get_login_type`. Use `loginType` property instead.",
            DeprecationWarning,
        )
        return self.loginType

    def is_user_login_type(self):
        return self.loginType == "user"

    def _authenticate(self, user_name, password):
        if user_name != "idtest0":
            raise Exception(f"Could not authenticate")

        if password == "wrong":  # noqa: S105
            raise Exception("Could not authenticate")

        if password == "ispybDown":  # noqa: S105
            raise Exception("Could not authenticate")

    def _create_test_session(self):
        session_dict = {
            "session_id": "1565334143",
            "beamline_name": "ID23-1",
            "start_date": "20240615",
            "start_time": "14:50:34",
            "end_date": "20240925",
            "end_time": "14:50:34",
            "title": "MXCuBE Sample tracking Development ",
            "code": "ID23-1",
            "number": "0424",
            "proposal_id": "1565334143",
            "proposal_name": "ID23-1-0424",
            "comments": "",
            "nb_shifts": "3",
            "scheduled": "True",
            "is_rescheduled": False,
            "is_scheduled_time": True,
            "is_scheduled_beamline": True,
            "user_portal_URL": "",
            "data_portal_URL": "https://data2.esrf.fr/investigation/1565334143/datasets",
            "logbook_URL": "https://data2.esrf.fr/investigation/1565334143/logbook",
        }

        session: Session = Session(**session_dict)
        return LimsSessionManager(sessions=[session], active_session=session)

    def login(
        self, user_name: str, password: str, is_local_host: bool
    ) -> LimsSessionManager:
        logging.getLogger("HRW").debug(
            "Login on ISPyBLims proposal=%s is_local_host=%s"
            % (user_name, str(is_local_host)),
        )
        self._authenticate(user_name, password)
        self.session_manager = LimsSessionManager()
        # Authentication
        try:
            self._authenticate(user_name, password)
            self.user_name = user_name
        except BaseException as e:
            raise e

        self.session_manager = self._create_test_session()
        return self.session_manager

    def create_session(self, proposal):
        self.session_manager = self._create_test_session()
        return self.session_manager

    def get_proposal(self, proposal_code, proposal_number):
        """
        Returns the tuple (Proposal, Person, Laboratory, Session, Status).
        Containing the data from the corresponding tables in the database
        the status of the database operations are returned in Status.

        :param proposal_code: The proposal code
        :type proposal_code: str
        :param proposal_number: The proposal number
        :type proposal_number: int

        :returns: The dict (Proposal, Person, Laboratory, Sessions, Status).
        :rtype: dict
        """
        return self.__test_proposal

    def get_proposals_by_user(self, user_name):
        return [self.__test_proposal]

    def get_session_local_contact(self, session_id):
        return {
            "personId": 1,
            "laboratoryId": 1,
            "login": None,
            "familyName": "operator on ID14eh1",
        }

    def translate(self, code, what):
        """
        Given a proposal code, returns the correct code to use in the GUI,
        or what to send to LDAP, user office database, or the ISPyB database.
        """
        try:
            translated = self.__translations[code][what]
        except KeyError:
            translated = code

        return translated

    def is_connected(self):
        return self.login_ok

    def isInhouseUser(self, proposal_code, proposal_number):
        """
        Returns True if the proposal is considered to be a
        in-house user.

        :param proposal_code:
        :type proposal_code: str

        :param proposal_number:
        :type proposal_number: str

        :rtype: bool
        """
        for proposal in self["inhouse"]:
            if proposal_code == proposal.code:
                if str(proposal_number) == str(proposal.number):
                    return True
        return False

    def _store_data_collection_group(self, group_data_dict):
        pass

    def store_data_collection(self, mx_collection, bl_config=None):
        """
        Stores the data collection mx_collection, and the beamline setup
        if provided.

        :param mx_collection: The data collection parameters.
        :type mx_collection: dict

        :param bl_config: The beamline setup.
        :type bl_config: dict

        :returns: None

        """
        logging.getLogger("HWR").debug(
            "Data collection parameters stored " + "in ISPyB: %s" % str(mx_collection)
        )
        logging.getLogger("HWR").debug(
            "Beamline setup stored in ISPyB: %s" % str(bl_config)
        )

        return None, None

    def store_beamline_setup(self, session_id, bl_config):
        """
        Stores the beamline setup dict <bl_config>.

        :param session_id: The session id that the bl_config
                           should be associated with.
        :type session_id: int

        :param bl_config: The dictionary with beamline settings.
        :type bl_config: dict

        :returns beamline_setup_id: The database id of the beamline setup.
        :rtype: str
        """
        pass

    def update_data_collection(self, mx_collection, wait=False):
        """
        Updates the datacollction mx_collection, this requires that the
        collectionId attribute is set and exists in the database.

        :param mx_collection: The dictionary with collections parameters.
        :type mx_collection: dict

        :returns: None
        """
        pass

    def update_bl_sample(self, bl_sample):
        """
        Creates or stos a BLSample entry.

        :param sample_dict: A dictionary with the properties for the entry.
        :type sample_dict: dict
        # NBNB update doc string
        """
        pass

    def store_image(self, image_dict):
        """
        Stores the image (image parameters) <image_dict>

        :param image_dict: A dictionary with image pramaters.
        :type image_dict: dict

        :returns: None
        """
        pass

    def store_robot_action(self, robot_action_dict):
        """
        Stores the robot action dictionary.

        Structure of robot_action_dictionary:
        {
            "actionType":str,
            "containerLocation": str,
            "dewarLocation":str,
            "message":str,
            "sampleBarcode":str,
            "sessionId":int,
            "sampleId":int.
            "startTime":str,
            "endTime":str,
            "xtalSnapshotAfter:str",
            "xtalSnapshotBefore:str",
        }

        Args:
            robot_action_dict: robot action dictionary as defined above
        """
        pass

    def __find_sample(self, sample_ref_list, code=None, location=None):
        """
        Returns the sample with the matching "search criteria" <code> and/or
        <location> with-in the list sample_ref_list.

        The sample_ref object is defined in the head of the file.

        :param sample_ref_list: The list of sample_refs to search.

        :param code: The vial datamatrix code (or bar code)

        :param location: A tuple (<basket>, <vial>) to search for.
        :type location: tuple
        """
        pass

    def get_samples(self, lims_name):
        # Try GPhL emulation samples, if available
        gphl_workflow = HWR.beamline.gphl_workflow
        if gphl_workflow is not None:
            sample_dicts = gphl_workflow.get_emulation_samples()
            if sample_dicts:
                return sample_dicts

        return [
            {
                "cellA": 0.0,
                "cellAlpha": 0.0,
                "cellB": 0.0,
                "cellBeta": 0.0,
                "cellC": 0.0,
                "cellGamma": 0.0,
                "containerSampleChangerLocation": "1",
                "crystalSpaceGroup": "P212121",
                "diffractionPlan": {
                    "diffractionPlanId": 457980,
                    "experimentKind": "Default",
                    "numberOfPositions": 0,
                    "observedResolution": 0.0,
                    "preferredBeamDiameter": 0.0,
                    "radiationSensitivity": 0.0,
                    "requiredCompleteness": 0.0,
                    "requiredMultiplicity": 0.0,
                    "requiredResolution": 0.0,
                },
                "experimentType": "Default",
                "proteinAcronym": "A-TIM",
                "sampleId": 515485,
                "sampleLocation": "1",
                "sampleName": "fghfg",
                "smiles": None,
            },
            {
                "cellA": 0.0,
                "cellAlpha": 0.0,
                "cellB": 0.0,
                "cellBeta": 0.0,
                "cellC": 0.0,
                "cellGamma": 0.0,
                "containerSampleChangerLocation": "2",
                "crystalSpaceGroup": "P2",
                "diffractionPlan": {
                    "diffractionPlanId": 457833,
                    "experimentKind": "OSC",
                    "numberOfPositions": 0,
                    "observedResolution": 0.0,
                    "preferredBeamDiameter": 0.0,
                    "radiationSensitivity": 0.0,
                    "requiredCompleteness": 0.0,
                    "requiredMultiplicity": 0.0,
                    "requiredResolution": 0.0,
                },
                "experimentType": "OSC",
                "proteinAcronym": "B2 hexa",
                "sampleId": 515419,
                "sampleLocation": "1",
                "sampleName": "sample",
            },
            {
                "cellA": 0.0,
                "cellAlpha": 0.0,
                "cellB": 0.0,
                "cellBeta": 0.0,
                "cellC": 0.0,
                "cellGamma": 0.0,
                "containerSampleChangerLocation": "2",
                "crystalSpaceGroup": "P2",
                "diffractionPlan": {
                    "diffractionPlanId": 457834,
                    "experimentKind": "OSC",
                    "numberOfPositions": 0,
                    "observedResolution": 0.0,
                    "preferredBeamDiameter": 0.0,
                    "radiationSensitivity": 0.0,
                    "requiredCompleteness": 0.0,
                    "requiredMultiplicity": 0.0,
                    "requiredResolution": 0.0,
                },
                "experimentType": "OSC",
                "proteinAcronym": "B2 hexa",
                "sampleId": 515420,
                "sampleLocation": "2",
                "sampleName": "sample",
            },
            {
                "cellA": 0.0,
                "cellAlpha": 0.0,
                "cellB": 0.0,
                "cellBeta": 0.0,
                "cellC": 0.0,
                "cellGamma": 0.0,
                "containerSampleChangerLocation": "2",
                "crystalSpaceGroup": "P2",
                "diffractionPlan": {
                    "diffractionPlanId": 457835,
                    "experimentKind": "OSC",
                    "numberOfPositions": 0,
                    "observedResolution": 0.0,
                    "preferredBeamDiameter": 0.0,
                    "radiationSensitivity": 0.0,
                    "requiredCompleteness": 0.0,
                    "requiredMultiplicity": 0.0,
                    "requiredResolution": 0.0,
                },
                "experimentType": "OSC",
                "proteinAcronym": "B2 hexa",
                "sampleId": 515421,
                "sampleLocation": "3",
                "sampleName": "sample",
            },
            {
                "cellA": 0.0,
                "cellAlpha": 0.0,
                "cellB": 0.0,
                "cellBeta": 0.0,
                "cellC": 0.0,
                "cellGamma": 0.0,
                "containerSampleChangerLocation": "2",
                "crystalSpaceGroup": "P2",
                "diffractionPlan": {
                    "diffractionPlanId": 457836,
                    "experimentKind": "OSC",
                    "numberOfPositions": 0,
                    "observedResolution": 0.0,
                    "preferredBeamDiameter": 0.0,
                    "radiationSensitivity": 0.0,
                    "requiredCompleteness": 0.0,
                    "requiredMultiplicity": 0.0,
                    "requiredResolution": 0.0,
                },
                "experimentType": "OSC",
                "proteinAcronym": "B2 hexa",
                "sampleId": 515422,
                "sampleLocation": "5",
                "sampleName": "sample",
            },
            {
                "cellA": 0.0,
                "cellAlpha": 0.0,
                "cellB": 0.0,
                "cellBeta": 0.0,
                "cellC": 0.0,
                "cellGamma": 0.0,
                "containerSampleChangerLocation": "2",
                "crystalSpaceGroup": "P2",
                "diffractionPlan": {
                    "diffractionPlanId": 457837,
                    "experimentKind": "OSC",
                    "numberOfPositions": 0,
                    "observedResolution": 0.0,
                    "preferredBeamDiameter": 0.0,
                    "radiationSensitivity": 0.0,
                    "requiredCompleteness": 0.0,
                    "requiredMultiplicity": 0.0,
                    "requiredResolution": 0.0,
                },
                "experimentType": "OSC",
                "proteinAcronym": "B2 hexa",
                "sampleId": 515423,
                "sampleLocation": "6",
                "sampleName": "sample",
            },
            {
                "cellA": 0.0,
                "cellAlpha": 0.0,
                "cellB": 0.0,
                "cellBeta": 0.0,
                "cellC": 0.0,
                "cellGamma": 0.0,
                "containerSampleChangerLocation": "2",
                "crystalSpaceGroup": "P2",
                "diffractionPlan": {
                    "diffractionPlanId": 457838,
                    "experimentKind": "OSC",
                    "numberOfPositions": 0,
                    "observedResolution": 0.0,
                    "preferredBeamDiameter": 0.0,
                    "radiationSensitivity": 0.0,
                    "requiredCompleteness": 0.0,
                    "requiredMultiplicity": 0.0,
                    "requiredResolution": 0.0,
                },
                "experimentType": "OSC",
                "proteinAcronym": "B2 hexa",
                "sampleId": 515424,
                "sampleLocation": "7",
                "sampleName": "sample",
            },
        ]

    # Bindings to methods called from older bricks.
    getProposal = get_proposal
    getSessionLocalContact = get_session_local_contact
    storeDataCollection = store_data_collection
    storeBeamLineSetup = store_beamline_setup
    updateBLSample = update_bl_sample
    updateDataCollection = update_data_collection
    storeImage = store_image



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""Beamline class serving as singleton container for links to top-level HardwareObjects

All HardwareObjects
"""

from __future__ import (
    absolute_import,
    division,
    print_function,
    unicode_literals,
)

from typing import (
    Any,
    Union,
)
from warnings import warn

from mxcubecore.dispatcher import dispatcher

__copyright__ = """ Copyright © 2019 by the MXCuBE collaboration """
__license__ = "LGPLv3+"
__author__ = "Rasmus H Fogh"

import logging

from mxcubecore.BaseHardwareObjects import (
    ConfiguredObject,
    HardwareObject,
)

# NBNB The acq parameter names match the attributes of AcquisitionParameters
# Whereas the limit parameter values use more understandable names
#
# TODO Make all tags consistent, including AcquisitionParameters attributes.


class Beamline(HardwareObject):
    """Beamline class serving as singleton container for links to HardwareObjects"""

    class HOConfig(ConfiguredObject.HOConfig):
        # Properties - definition and default values

        # List[str] of advanced method names
        advanced_methods = []

        # List[str] of available methods
        available_methods = []

        # int number of clicks used for click centring
        click_centring_num_clicks = 3

        # bool Is wavelength tunable
        tunable_wavelength = False

        # bool Disable number-of-passes widget NBNB TODO Move elsewhere??
        disable_num_passes = False

        # bool By default run online processing (characterization/mesh?)
        run_online_processing = False

        offline_processing_methods = []

        online_processing_methods = []

        # Dictionary-of-dictionaries of default acquisition parameters
        default_acquisition_parameters = {}

        # int Starting run number for path_template
        run_number = 1

        # List of undulators
        undulators = []

        # Format of mesh result for display
        mesh_result_format = "PNG"

        # bool Use the native mesh feature available, true by default
        use_native_mesh = True

        # bool Enable features to work with points in the plane, called
        # 2D-points, (none centred positions)
        enable_2d_points = True

        # Contained hardware objects

    @property
    def machine_info(self) -> HardwareObject | None:
        return self.get_object_by_role("machine_info")

    @property
    def transmission(self) -> HardwareObject | None:
        return self.get_object_by_role("transmission")

    @property
    def cryo(self) -> HardwareObject | None:
        return self.get_object_by_role("cryo")

    @property
    def energy(self) -> HardwareObject | None:
        return self.get_object_by_role("energy")

    @property
    def flux(self) -> HardwareObject | None:
        return self.get_object_by_role("flux")

    @property
    def beam(self) -> HardwareObject | None:
        return self.get_object_by_role("beam")

    @property
    def hutch_interlock(self) -> HardwareObject | None:
        return self.get_object_by_role("hutch_interlock")

    @property
    def safety_shutter(self) -> HardwareObject | None:
        return self.get_object_by_role("safety_shutter")

    @property
    def fast_shutter(self) -> HardwareObject | None:
        return self.get_object_by_role("fast_shutter")

    @property
    def diffractometer(self) -> HardwareObject | None:
        return self.get_object_by_role("diffractometer")

    @property
    def detector(self) -> HardwareObject | None:
        return self.get_object_by_role("detector")

    @property
    def resolution(self) -> HardwareObject | None:
        return self.get_object_by_role("resolution")

    @property
    def sample_changer(self) -> HardwareObject | None:
        return self.get_object_by_role("sample_changer")

    @property
    def sample_changer_maintenance(self) -> HardwareObject | None:
        return self.get_object_by_role("sample_changer_maintenance")

    @property
    def harvester(self) -> HardwareObject | None:
        return self.get_object_by_role("harvester")

    @property
    def harvester_maintenance(self) -> HardwareObject | None:
        return self.get_object_by_role("harvester_maintenance")

    @property
    def plate_manipulator(self) -> HardwareObject | None:
        return self.get_object_by_role("plate_manipulator")

    @property
    def session(self) -> HardwareObject | None:
        return self.get_object_by_role("session")

    @property
    def lims(self) -> HardwareObject | None:
        return self.get_object_by_role("lims")

    @property
    def sample_view(self) -> HardwareObject | None:
        return self.get_object_by_role("sample_view")

    @property
    def queue_manager(self) -> HardwareObject | None:
        return self.get_object_by_role("queue_manager")

    @property
    def queue_model(self) -> HardwareObject | None:
        return self.get_object_by_role("queue_model")

    @property
    def collect(self) -> HardwareObject | None:
        return self.get_object_by_role("collect")

    @property
    def xrf_spectrum(self) -> HardwareObject | None:
        return self.get_object_by_role("xrf_spectrum")

    @property
    def energy_scan(self) -> HardwareObject | None:
        return self.get_object_by_role("energy_scan")

    @property
    def imaging(self) -> HardwareObject | None:
        return self.get_object_by_role("imaging")

    @property
    def beamline_actions(self) -> HardwareObject | None:
        return self.get_object_by_role("beamline_actions")

    @property
    def xml_rpc_server(self) -> HardwareObject | None:
        return self.get_object_by_role("xml_rpc_server")

    @property
    def workflow(self) -> HardwareObject | None:
        return self.get_object_by_role("workflow")

    @property
    def control(self) -> HardwareObject | None:
        return self.get_object_by_role("control")

    @property
    def gphl_workflow(self) -> HardwareObject | None:
        return self.get_object_by_role("gphl_workflow")

    @property
    def gphl_connection(self) -> HardwareObject | None:
        return self.get_object_by_role("gphl_connection")

    @property
    def xray_centring(self) -> HardwareObject | None:
        return self.get_object_by_role("xray_centring")

    @property
    def online_processing(self) -> HardwareObject | None:
        return self.get_object_by_role("online_processing")

    @property
    def offline_processing(self) -> HardwareObject | None:
        return self.get_object_by_role("offline_processing")

    @property
    def characterisation(self) -> HardwareObject | None:
        return self.get_object_by_role("characterisation")

    @property
    def image_tracking(self) -> HardwareObject | None:
        return self.get_object_by_role("image_tracking")

    @property
    def procedure(self) -> HardwareObject | None:
        return self.get_object_by_role("procedure")

    @property
    def data_publisher(self) -> HardwareObject | None:
        return self.get_object_by_role("data_publisher")

    def _init(self) -> None:
        """Object initialisation - executed *before* loading contents"""

    def init(self):
        """Object initialisation - executed *after* loading contents"""

    def _hwr_init_done(self):
        """
        Method called after the initialization of HardwareRepository is done
        (when all HardwareObjects have been created and initialized)
        """
        self._hardware_object_id_dict = self._get_id_dict()

    def get_id(self, ho: HardwareObject) -> str:
        warn("Beamline.get_id is Deprecated. Use hwobj.id instead", stacklevel=2)
        return ho.id

    def get_hardware_object(self, _id: str) -> Union[HardwareObject, None]:
        warn(
            "Beamline.get_hardware_object is Deprecated. Use get_by_id instead",
            stacklevel=2,
        )
        return self.get_by_id(_id)

    # Signal handling functions:
    def emit(self, signal: Union[str, object, Any], *args) -> None:
        """Emit signal. Accepts both multiple args and a single tuple of args.

        This is needed for communication from the GUI to the core
        (jsonparamsgui in mxcubeqt)

        NBNB TODO HACK
        This is a duplicate of the same function in HardwareObjectMixin.
        Since the Beamline is not a CommandContainer or a normal HardwareObject
        it may not be appropriate to make it a subclass of HardwareObjectYaml
        We need to consider how we want this organised

        Args:
            signal (Union[str, object, Any]): In practice a string, or dispatcher.
            *args (tuple): Arguments sent with signal.
        """

        signal = str(signal)

        if len(args) == 1:
            if isinstance(args[0], tuple):
                args = args[0]
        responses: list = dispatcher.send(signal, self, *args)
        if not responses:
            raise RuntimeError("Signal %s is not connected" % signal)

    # Additional functions

    # NB Objects need not be HardwareObjects
    # We still categorise them as'hardware' if they are not procedures, though
    # The attribute values will be given in the config.yml file
    def get_default_acquisition_parameters(self, acquisition_type="default"):
        """
        :returns: A AcquisitionParameters object with all default parameters for the
                  specified acquisition type. "default" is a standard acqquisition
        """
        # Imported here to avoid circular imports
        from mxcubecore.model import queue_model_objects

        acq_parameters = queue_model_objects.AcquisitionParameters()

        params = self.config.default_acquisition_parameters["default"].copy()
        if acquisition_type != "default":
            dd0 = self.config.default_acquisition_parameters.get(acquisition_type)
            if dd0 is None:
                logging.getLogger("HWR").warning(
                    "No separate parameters for acquisition type: %s - using default."
                    % acquisition_type
                )
            else:
                params.update(dd0)

        for tag, val in params.items():
            setattr(acq_parameters, tag, val)

        motor_positions = self.diffractometer.get_positions()
        osc_start = motor_positions.get("phi")
        if osc_start is None:
            acq_parameters.osc_start = params.get("osc_start")
        else:
            acq_parameters.osc_start = osc_start

        kappa = motor_positions.get("kappa")
        if kappa is None:
            acq_parameters.kappa = None
        else:
            acq_parameters.kappa = round(float(kappa), 2)

        kappa_phi = motor_positions.get("kappa_phi")
        if kappa_phi is None:
            acq_parameters.kappa_phi = None
        else:
            acq_parameters.kappa_phi = round(float(kappa_phi), 2)

        try:
            acq_parameters.resolution = self.resolution.get_value()
        except Exception:
            logging.getLogger("HWR").warning(
                "get_default_acquisition_parameters: "
                "No current resolution, setting to 0.0"
            )
            acq_parameters.resolution = 0.0

        try:
            acq_parameters.energy = self.energy.get_value()
        except Exception:
            logging.getLogger("HWR").warning(
                "get_default_acquisition_parameters: No current energy, setting to 0.0"
            )
            acq_parameters.energy = 0.0

        try:
            acq_parameters.transmission = self.transmission.get_value()
        except Exception:
            logging.getLogger("HWR").warning(
                "get_default_acquisition_parameters: "
                "No current transmission, setting to 0.0"
            )
            acq_parameters.transmission = 0.0

        acq_parameters.shutterless = params.get("shutterless", True)

        try:
            acq_parameters.detector_binning_mode = self.detector.get_binning_mode()
        except Exception:
            logging.getLogger("HWR").warning(
                "get_default_acquisition_parameters: "
                "Could not get detector mode, setting to ''"
            )
            acq_parameters.detector_binning_mode = ""

        try:
            acq_parameters.detector_roi_mode = self.detector.get_roi_mode()
        except Exception:
            logging.getLogger("HWR").warning(
                "get_default_acquisition_parameters: "
                "Could not get roi mode, setting to ''"
            )
            acq_parameters.detector_roi_mode = ""

        return acq_parameters

    def get_default_path_template(self):
        """
        :returns: A PathTemplate object with default parameters.
        """
        # Imported here to avoid circular imports
        from mxcubecore.model import queue_model_objects

        path_template = queue_model_objects.PathTemplate()

        acq_params = self.get_default_acquisition_parameters()
        path_template.start_num = acq_params.first_image
        path_template.num_files = acq_params.num_images

        path_template.run_number = self.config.run_number

        return path_template

    def get_default_characterisation_parameters(self):
        return self.characterisation.get_default_characterisation_parameters()

    def force_emit_signals(self):
        hwobjs = list(self.objects_by_role.values())
        for hwobj in hwobjs:
            hwobj.force_emit_signals()
            hwobjs.extend(hwobj.objects_by_role.values())



#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.
#
#  Please user PEP 0008 -- "Style Guide for Python Code" to format code
#  https://www.python.org/dev/peps/pep-0008/

"""
Handles interaction with the data model(s). Adding, removing and
retrieving nodes are all done via this object. It is possible to
handle several models by using register_model and select_model.
"""

import json
import logging

import jsonpickle

from mxcubecore import HardwareRepository as HWR
from mxcubecore import queue_entry
from mxcubecore.BaseHardwareObjects import HardwareObject
from mxcubecore.model import queue_model_objects


class Serializer(object):
    @staticmethod
    def serialize(object):
        return json.dumps(object, default=lambda o: o.__dict__.values()[0])


class QueueModel(HardwareObject):
    def __init__(self, name):
        HardwareObject.__init__(self, name)

        self._ispyb_model = queue_model_objects.RootNode()
        self._ispyb_model._node_id = 0
        self._free_pin_model = queue_model_objects.RootNode()
        self._free_pin_model._node_id = 0
        self._plate_model = queue_model_objects.RootNode()
        self._plate_model._node_id = 0

        self._models = {
            "ispyb": self._ispyb_model,
            "free-pin": self._free_pin_model,
            "plate": self._plate_model,
        }

        self._selected_model = self._ispyb_model

    def __getstate__(self):
        d = dict(self.__dict__)
        return d

    def __setstate__(self, d):
        self.__dict__.update(d)

    # Framework-2 method, inherited from HardwareObject and called
    # by the framework after the object has been initialized.
    def init(self):
        """
        Framework-2 method, inherited from HardwareObject and called
        by the framework after the object has been initialized.

        You should normally not need to call this method.
        """
        pass

    def select_model(self, name):
        """
        Selects the model with the name <name>

        :param name: The name of the model to select.
        :type name: str

        :returns: None
        :rtype: NoneType
        """
        self._selected_model = self._models[name]
        HWR.beamline.queue_manager.clear()
        self._re_emit(self._selected_model)

    def get_model_root(self):
        """
        :returns: The selected model root.
        :rtype: TaskNode
        """
        return self._selected_model

    def clear_model(self, name=None):
        """
        Clears the model with name <name>, clears all if name is None

        :param name: The name of the model to clear.
        :type name: str

        :returns: None
        :rtype: NoneType
        """
        self._models[name] = queue_model_objects.RootNode()

        if not name:
            for name in self._models.keys():
                self._models[name] = queue_model_objects.RootNode()

        HWR.beamline.queue_manager.clear()

    def register_model(self, name, root_node):
        """
        Register a new model with name <name> and root node <root_node>.

        :param name: The name of the 'new' model.
        :type name: str

        :param root_node: The root of the model.
        :type root_node: RootNode

        :returns: None
        :rtype: NoneType
        """
        if name in self._models:
            raise KeyError("The key %s is already registered" % name)
        else:
            self._models[name]

    def _re_emit(self, parent_node):
        """
        Re-emits the 'child_added' for all the nodes in the model.
        """
        for child_node in parent_node.get_children():
            self.emit("child_added", (parent_node, child_node))
            self._re_emit(child_node)

    def add_child(self, parent, child):
        """
        Adds the child node <child>. Raises the exception TypeError
        if child is not of type TaskNode.

        Moves the child (re-parents it) if it already has a parent.

        :param child: TaskNode to add
        :type child: TaskNode

        :returns: None
        :rtype: None
        """
        if True:
            # if isinstance(child, queue_model_objects.TaskNode):
            self._selected_model._total_node_count += 1
            child._parent = parent
            child._node_id = self._selected_model._total_node_count
            parent._children.append(child)
            child._set_name(child._name)
            self.emit("child_added", (parent, child))
        else:
            raise TypeError("Expected type TaskNode, got %s " % str(type(child)))

    def add_child_at_id(self, _id, child):
        """
        Adds a child <child> at the node with the node id <_id>

        :param _id: The id of the parent node.
        :type _id: int

        :param child: The child node to add.
        :type child: TaskNode

        :returns: The id of the child.
        :rtype: int
        """
        parent = self.get_node(_id)
        self.add_child(parent, child)
        return child._node_id

    def get_node(self, _id, parent=None):
        """
        Retrieves the node with the node id <_id>

        :param _id: The id of the node to retrieve.
        :type _id: int

        :param parent: parent node to search in.
        :type parent: TaskNode

        :returns: The node with the id <_id>
        :rtype: TaskNode
        """
        if parent is None:
            parent = self._selected_model

        for node in parent._children:
            if node._node_id == _id:
                return node
            else:
                result = self.get_node(_id, node)

                if result:
                    return result

    def del_child(self, parent, child):
        """
        Removes <child>

        :param child: Child to remove.
        :type child: TaskNode

        :returns: None
        :rtype: None
        """
        if child in parent._children:
            parent._children.remove(child)
            self.emit("child_removed", (parent, child))

    def _detach_child(self, parent, child):
        """
        Detaches the child <child>

        :param child: Child to detach.
        :type child: TaskNode

        :returns: None
        :rtype: None
        """
        child = parent._children.pop(child)
        return child

    def set_parent(self, parent, child):
        """
        Sets the parent of the child <child> to <parent>

        :param parent: The parent.
        :type parent: TaskNode Object

        :param child: The child
        :type child: TaskNode Object
        """
        if child._parent:
            self._detach_child(parent, child)
            child.set_parent(parent)
        else:
            child._parent = parent

    def view_created(self, view_item, task_model):
        """
        Method that should be called by the routine that adds
        the view <view_item> for the model <task_model>

        :param view_item: The view item that was added.
        :type view_item: ViewItem

        :param task_model: The associated task model.
        :type task_model: TaskModel

        :returns: None
        :rtype: None
        """
        view_item._data_model = task_model
        cls = queue_entry.MODEL_QUEUE_ENTRY_MAPPINGS[task_model.__class__]
        qe = cls(view_item, task_model)
        # view_item.setText(0, task_model.get_name())

        # if isinstance(task_model, queue_model_objects.Sample) or \
        #  isinstance(task_model, queue_model_objects.TaskGroup):
        #    view_item.setText(0, task_model.get_name())
        # else:
        view_item.setText(0, task_model.get_display_name())

        view_item.setOn(task_model.is_enabled())

        if isinstance(task_model, queue_model_objects.Sample):
            HWR.beamline.queue_manager.enqueue(qe)
        elif not isinstance(task_model, queue_model_objects.Basket):
            # else:
            view_item.parent().get_queue_entry().enqueue(qe)
        view_item.update_tool_tip()

    def get_next_run_number(self, new_path_template, exclude_current=True):
        """
        Iterates through all the path templates of the tasks
        in the model and returns the next available run number
        for the path template <new_path_template>.

        :param new_path_template: PathTempalte to match with.
        :type new_path_template: PathTemplate
        :param exclude_current: Skips it self when iterating through
                                the model, default Tree.
        :type exlcude_current: bool

        :returns: The next available run number for the given path_template.
        :rtype: int
        """
        all_path_templates = self.get_path_templates()
        conflicting_path_templates = [0]

        for pt in all_path_templates:
            if exclude_current:
                if pt[1] is not new_path_template:
                    if pt[1] == new_path_template:
                        conflicting_path_templates.append(pt[1].run_number)
            else:
                if pt[1] == new_path_template:
                    conflicting_path_templates.append(pt[1].run_number)

        return max(conflicting_path_templates) + 1

    def get_path_templates(self):
        """
        Retrieves a list of all the path templates in the model.
        """
        return self._get_path_templates_rec(self.get_model_root())

    def _get_path_templates_rec(self, parent_node):
        """
        Recursive part of get_path_templates.
        """
        path_template_list = []

        for child_node in parent_node.get_children():
            path_template = child_node.get_path_template()

            if path_template:
                path_template_list.append((child_node, path_template))

            child_path_template_list = self._get_path_templates_rec(child_node)

            if child_path_template_list:
                path_template_list.extend(child_path_template_list)

        return path_template_list

    def check_for_path_collisions(self, new_path_template):
        """
        Returns True if there is a path template (task) in the model,
        that produces the same files as this one.

        :returns: True if there is a potential path collision.
        """
        result = False
        path_template_list = self.get_path_templates()

        for pt in path_template_list:
            if pt[1] is not new_path_template:
                if new_path_template.intersection(pt[1]):
                    result = True

        return result

    def copy_node(self, node):
        """
        Copies the node <node> and returns it.

        :param node: The node to copy.
        :type node: TaskModel

        :returns: A copy of the node.
        :rtype: TaskModel
        """
        new_node = node.copy()

        if new_node.get_path_template():
            pt = new_node.get_path_template()
            new_run_number = self.get_next_run_number(pt)
            pt.run_number = new_run_number
            new_node.set_number(new_run_number)

        # We do not copy grid object, but keep a link to the original grid
        if hasattr(new_node, "grid"):
            new_node.grid = node.grid

        new_node.set_executed(False)

        return new_node

    def get_nodes(self):
        node_list = []

        def get_nodes_list(entry):
            for child in entry._children:
                node_list.append(child)
                get_nodes_list(child)

        for qe in self._selected_model._children:
            get_nodes_list(qe)

        return node_list

    def get_all_queue_entries(self):
        node_list = []

        def get_nodes_list(entry):
            for child in entry._queue_entry_list:
                node_list.append(child)
                get_nodes_list(child)

        for qe in HWR.beamline.queue_manager._queue_entry_list:
            get_nodes_list(qe)

        return node_list

    def get_all_dc_queue_entries(self):
        result = []

        for item in self.get_all_queue_entries():
            if isinstance(item, queue_entry.DataCollectionQueueEntry):
                result.append(item)

        return result

    def save_queue(self, filename=None):
        """Saves queue in the file. Current selected model is saved as a list
        of dictionaries. Information about samples and baskets is not saved
        """
        if not filename:
            # filename = os.path.join(self.user_file_directory, "queue_active.dat")
            filename = "queue_active.dat"

        items_to_save = []

        selected_model = ""
        for key in self._models:
            if self._selected_model == self._models[key]:
                selected_model = key

        queue_entry_list = HWR.beamline.queue_manager.get_queue_entry_list()
        for item in queue_entry_list:
            # On the top level is Sample or Basket
            if isinstance(item, queue_entry.SampleQueueEntry):
                for task_item in item.get_queue_entry_list():
                    task_item_dict = {
                        "sample_location": item.get_data_model().location,
                        "task_group_entry": jsonpickle.encode(
                            task_item.get_data_model()
                        ),
                    }
                    items_to_save.append(task_item_dict)

        save_file = None
        try:
            save_file = open(filename, "w")
            save_file.write(repr((selected_model, items_to_save)))
        except Exception:
            logging.getLogger().exception(
                "Unable to save queue " + "in file %s", filename
            )
            if save_file:
                save_file.close()

    def get_queue_as_json_list(self):
        items_to_save = []

        selected_model = ""
        for key in self._models:
            if self._selected_model == self._models[key]:
                selected_model = key

        queue_entry_list = HWR.beamline.queue_manager.get_queue_entry_list()
        for item in queue_entry_list:
            # On the top level is Sample or Basket
            if isinstance(item, queue_entry.SampleQueueEntry):
                for task_item in item.get_queue_entry_list():
                    task_item_dict = {
                        "sample_location": item.get_data_model().location,
                        # "task_group_entry": Serializer.serialize(task_item.get_data_model())}
                        # "task_group_entry" : jsonpickle.encode(task_item.get_data_model())}
                        "task_group_entry": json.dumps(task_item.get_data_model()),
                    }
                    items_to_save.append(task_item_dict)

        return selected_model, items_to_save

    def load_queue_from_json_list(self, queue_list, snapshot):
        # Prepare list of samplesL
        sample_dict = {}
        for item in HWR.beamline.queue_manager.get_queue_entry_list():
            if isinstance(item, queue_entry.SampleQueueEntry):
                sample_data_model = item.get_data_model()
                sample_dict[sample_data_model.location] = sample_data_model
            elif isinstance(item, queue_entry.BasketQueueEntry):
                for sample_item in item.get_queue_entry_list():
                    sample_data_model = sample_item.get_data_model()
                    sample_dict[sample_data_model.location] = sample_data_model
        if len(queue_list) > 0:
            try:
                for task_group_item in queue_list:
                    task_group_entry = json.load(task_group_item["task_group_entry"])
                    self.add_child(
                        sample_dict[task_group_item["sample_location"]],
                        task_group_entry,
                    )
                    for child in task_group_entry.get_children():
                        child.set_snapshot(snapshot)
                logging.getLogger("HWR").info("Queue loading done")
            except Exception:
                logging.getLogger("HWR").exception("Unable to load queue")

    def load_queue_from_file(self, filename, snapshot=None):
        """Loads queue from file. The problem is snapshots that are
        not stored in the file, so we have to add new ones in
        the loading process

        :returns: model name 'free-pin', 'ispyb' or 'plate'
        """

        logging.getLogger("HWR").info("Loading queue from file %s" % filename)
        load_file = None
        try:
            # Read file and clear the model
            load_file = open(filename, "r")
            decoded_file = eval(load_file.read())
            self.select_model(decoded_file[0])

            # Prepare list of samples
            sample_dict = {}
            for item in HWR.beamline.queue_manager.get_queue_entry_list():
                if isinstance(item, queue_entry.SampleQueueEntry):
                    sample_data_model = item.get_data_model()
                    sample_dict[sample_data_model.location] = sample_data_model
                elif isinstance(item, queue_entry.BasketQueueEntry):
                    for sample_item in item.get_queue_entry_list():
                        sample_data_model = sample_item.get_data_model()
                        sample_dict[sample_data_model.location] = sample_data_model

            if len(decoded_file[1]) > 0:
                for task_group_item in decoded_file[1]:
                    task_group_entry = jsonpickle.decode(
                        task_group_item["task_group_entry"]
                    )
                    self.add_child(
                        sample_dict[task_group_item["sample_location"]],
                        task_group_entry,
                    )
                    for child in task_group_entry.get_children():
                        child.set_snapshot(snapshot)
                logging.getLogger("HWR").info("Queue loading done")
            else:
                logging.getLogger("HWR").info("No queue content available in file")
            return decoded_file[0]
        except Exception:
            logging.getLogger("HWR").exception(
                "Unable to load queue " + "from file %s", filename
            )
            if load_file:
                load_file.close()



from datetime import (
    datetime,
    timedelta,
)
from typing import (
    Any,
    Dict,
    List,
    Optional,
)

from pydantic.v1 import (
    BaseModel,
    Field,
)


class Proposal(BaseModel):
    """Proposal with key identifying details.

    Parameters:
        proposal_id: Unique identifier for the proposal.
        person_id: Identifier of the person associated.
        type: Type/category of the proposal (e.g., "MX").
        code: Specific code associated with the proposal e.g., "IH-LS".
        number: Proposal number; may be sequential or unique
            (uniqueness may depend on the combination of code and number)
        title:  Title of the proposal
    """

    proposal_id: str = ""
    person_id: str = ""
    type: str = ""
    code: str = ""
    number: str = ""
    title: str = ""


class Lims(BaseModel):
    """Represents a LIMS system.

    Parameters:
        name: Unique identifier for the LIMS (e.g., "ispyb", "icat").
        description: Free text providing a short description of the LIMS.
    """

    name: str = ""
    description: str = ""


class Session(BaseModel):
    """Represents an experiment in a specific time slot on a beamline for a proposal.

    This class is maintained for legacy purposes because ISPyB relies on it.

    Parameters:
        session_id: Unique identifier for the experiment, dependent on the LIMS.
        beamline_name: Name of the beamline where the experiment is scheduled.
        start_date: Official start date (format: YYYYMMDD).
        start_time: Official start time.
        end_date: Official end date (format: YYYYMMDD).
        end_time: Official end time.
        title: Title of the proposal associated with the session.
        code: Code associated with the proposal.
        number: Proposal number.
        proposal_id: Unique identifier of the proposal.
        proposal_name: Name of the proposal.
        comments: Optional comments about the session.
        start_datetime: Start datetime of the session.
        end_datetime: End datetime of the session
        actual_start_date: Start date if the session was rescheduled.
        actual_start_time: Start time if the session was rescheduled.
        actual_end_date: End date if the session was rescheduled.
        actual_end_time: End time if the session was rescheduled.
        nb_shifts: Number of shifts allocated to the session (typically 8h)
        scheduled: if the session is officially scheduled and approved.
        is_rescheduled: if the session was rescheduled in time or beamline.
        is_scheduled_beamline: if the session is scheduled on the current bm
        is_scheduled_time: if the session is currently active.
        user_portal_URL: Optional link to the session page in the User Portal.
        data_portal_URL: link to the data portal or LIMS.
        logbook_URL: Optional link to the session page in the  logbook.
        volume: Optional volume (in bytes) of data produced.
        dataset_count: Optional number of datasets collected.
        sample_count: Optional number of samples collected.
    """

    session_id: str = ""
    beamline_name: str = ""
    start_date: str = ""
    start_time: str = ""
    end_date: str = ""
    end_time: str = ""
    title: str = ""
    code: str = ""
    number: str = ""
    proposal_id: str = ""
    proposal_name: str = ""
    comments: Optional[str] = ""
    start_datetime: datetime = Field(default_factory=datetime.now)
    end_datetime: Optional[datetime] = Field(
        default_factory=lambda: datetime.now() + timedelta(days=1)
    )
    actual_start_date: str = ""
    actual_start_time: str = ""
    actual_end_date: str = ""
    actual_end_time: str = ""
    nb_shifts: str = ""
    scheduled: str = ""
    is_rescheduled: bool = False
    is_scheduled_beamline: bool = False
    is_scheduled_time: bool = False
    user_portal_URL: Optional[str] = None
    data_portal_URL: Optional[str] = None
    logbook_URL: Optional[str] = None
    volume: Optional[str] = None
    dataset_count: Optional[str] = None
    sample_count: Optional[str] = None


class Instrument(BaseModel):
    """Represents a beamline.

    Parameters:
        name: Name of the beamline.
        id: Unique identifier of the beamline.
        instrumentScientists: List of scientists associated with the instrument.
    """

    name: str
    id: int
    instrumentScientists: List[Any]


class Investigation(BaseModel):
    """Represents an investigation and serves as a proposal to replace
    the Session class.

    Parameters:
        name: Name of the investigation.
        startDate: Start date of the investigation.
        endDate: End date of the investigation.
        id: Unique identifier of the investigation.
        title: Title of the investigation.
        visitId: Identifier for the visit associated with the investigation.
        summary: Summary description of the investigation.
        parameters: Dictionary of investigation-specific parameters.
        instrument: Associated instrument for the investigation.
        investigationUsers: List of users involved in the investigation.
    """

    name: str
    startDate: datetime
    endDate: datetime
    id: int
    title: str
    visitId: str
    summary: str
    parameters: Dict[str, Any]
    instrument: Instrument
    investigationUsers: List[Any]


class Parameter(BaseModel):
    """Represents a generic parameter attached to entities like investigations,
    samples, or datasets.

    Parameters:
        name: Name of the parameter.
        value: Value assigned to the parameter.
        id: Unique identifier of the parameter.
        units: Measurement units, if applicable.
    """

    name: str
    value: str
    id: int
    units: str


class MetaPage(BaseModel):
    """Pagination metadata.

    Parameters:
        totalWithoutFilters: Total count of items without filters applied.
        total: Total count of items with filters applied.
        totalPages: Total number of pages available.
        currentPage: Current page number.
    """

    totalWithoutFilters: int
    total: int
    totalPages: int
    currentPage: int


class Meta(BaseModel):
    """Metadata containing pagination details.

    Parameters:
        page: Pagination details.
    """

    page: MetaPage


class LimsUser(BaseModel):
    """Represents a user connected to MXCuBE.

    Parameters:
        user_name: Identifier for the user, typically their login name.
        sessions: List of sessions the user is allowed to collect data from.
    """

    user_name: str = ""
    sessions: Optional[List[Session]] = []


class LimsSessionManager(BaseModel):
    """Manages LIMS sessions and connected users.

    Parameters:
        active_session: The current active session, if any.
        sessions: List of selectable sessions determined based on connected users.
        users: Dictionary of currently connected users.
    """

    active_session: Optional[Session] = None
    sessions: Optional[List[Session]] = []
    users: Optional[Dict[str, LimsUser]] = {}


class SampleSheet(BaseModel):
    """Represents a description of a sample sheet as defined in user portals.

    Parameters:
        id: Unique identifier for the sample sheet.
        name: Name of the sample sheet, often corresponding to the protein's name.
        investigation: Investigation associated with the sample sheet.
        modTime: Last modification time of the sample sheet.
        parameters: Generic list of parameters, dependent on the user portal.
        datasets: List of datasets collected for this sample.
        meta: Pagination metadata.
    """

    id: int
    name: str
    investigation: Investigation
    modTime: datetime
    parameters: List[Parameter]
    datasets: List[Any]
    meta: Meta


class Resource(BaseModel):
    """
    Represents a resource, usually a file associated with a sample or
        investigation.

    Parameters:
        id (str): The unique identifier for the resource. The key "_id" is
            used in JSON and mapped to "id" in the class.
        filename (str): The name of the file associated with the resource.
        file (str): The URL or path pointing to the actual file or resource
            content.
        fileType (str): The type of the file (e.g., "pdb", "hkl", etc.).
        createdAt (datetime): The timestamp representing when the resource was
            created.
        updatedAt (datetime): The timestamp representing when the resource was
            last updated.
        groupName (Optional[str]): The optional group name associated with the
            resource (default is None).
    """

    id: str = Field(..., alias="_id")  # Use alias to map _id from JSON to id
    filename: str
    file: str
    fileType: str
    createdAt: datetime
    updatedAt: datetime
    groupName: Optional[str] = None


class SampleInformation(BaseModel):
    """
    Parameters metadata related to a sample, including a list of associated resources.

    Parameters:
        id (str): The unique identifier for the sample. The key "_id" is used
            in JSON and mapped to "id" in the class.
        investigationId (int): The unique identifier of the investigation this
            sample is associated with.
        sampleId (int): The unique identifier for the sample in ICAT. Most of the cases
            this represents the protein itself
        createdAt (datetime): The timestamp representing when the sample
            information was created.
        updatedAt (datetime): The timestamp representing when the sample
            information was last updated.
        resources (List[Resource]): A list of Resource objects associated
            with this sample information.
    """

    id: str = Field(..., alias="_id")  # Use alias to map _id from JSON to id
    investigationId: int
    sampleId: int
    createdAt: datetime
    updatedAt: datetime
    resources: List[Resource]  # List of associated resources


class Download(BaseModel):
    """
    Represents a downloadable file with metadata used to match pipeline references.

    Parameters:
        path: Absolute file path to the downloaded file on the filesystem.
        filename: Name of the file, used for matching with pipeline references.
        groupName:
            Optional group identifier, used to match with search models.
            Can be None or a string such as 'Group1', 'Group2', etc.
    """

    path: str = Field(..., description="Full path to the file on disk")
    filename: str = Field(..., description="Name of the file (e.g., 'example.hkl')")
    groupName: Optional[str] = Field(
        None, description="Group name used for model matching (e.g., 'Group1')"
    )



from typing import (
    Dict,
    Iterable,
    Optional,
    Tuple,
)

from pydantic import (
    BaseModel,
    RootModel,
)


class Command(BaseModel):
    """Sardana command configuration."""

    # name of the Sardana command
    name: Optional[str] = None


class Door(BaseModel):
    """Configuration of a Sardana door."""

    commands: Optional[Dict[str, Optional[Command]]] = None

    def get_commands(self) -> Iterable[Tuple[str, Command]]:
        """Get all commands configured for this door.

        This method will fill in optional configuration properties for commands.
        """
        if self.commands is None:
            return []

        for command_name, command_config in self.commands.items():
            if command_config is None:
                command_config = Command()  # noqa: PLW2901

            if command_config.name is None:
                command_config.name = command_name

            yield command_name, command_config


class SardanaConfig(RootModel[Dict[str, Door]]):
    """The 'sardana' section of the hardware object's YAML configuration file."""

    def get_doors(self) -> Iterable[Tuple[str, Door]]:
        """Get all Sardana doors specified in this section."""
        return list(self.root.items())



from mxcubecore.HardwareObjects import MultiplePositions


class MinidiffAperture(MultiplePositions.MultiplePositions):
    def __init__(self, *args, **kwargs):
        MultiplePositions.MultiplePositions.__init__(self, *args, **kwargs)

    def getApertureCoef(self):
        current_pos = self.get_value()
        for position in self["positions"]:
            if position.get_property("name") == current_pos:
                aperture_coef = position.get_property("aperture_coef")
                return aperture_coef
        return 1

    def getApertureSize(self):
        current_pos = self.get_value()
        for position in self["positions"]:
            if position.get_property("name") == current_pos:
                aperture_size = float(position.get_property("aperture_size"))

                if aperture_size > 1:
                    # aperture size in microns
                    return (aperture_size / 1000.0, aperture_size / 1000.0)
                else:
                    # aperture size in millimeters
                    return (aperture_size, aperture_size)
        return (9999, 9999)

    def connect_notify(self, signal):
        if signal == "apertureChanged":
            self.checkPosition()

    def checkPosition(self, *args):
        pos = MultiplePositions.MultiplePositions.checkPosition(self, *args)

        self.emit("apertureChanged", (self.getApertureSize(),))

        return pos



#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""
[Name]
QtTangoLimaVideo

[Description]
HwObj used to grab images via Tango Lima device server
If you want to access the Lima Library directly you may consider using
the QtLimaVideo module instead

[Configuration]
Example Hardware Object XML file :
==================================
<object class="QtLimaVideo">
   <type>basler</type>
   <encoding>yuv422p</encoding>
   <tangoname>bl13/eh/lima_oav</tangoname>
   <gain>0.5</gain>
   <exposure>0.01</exposure>
   <mirror>(False, False)</mirror>
   <interval>30</interval>
</object>
"""

import struct

import numpy as np
import PyTango

from mxcubecore.HardwareObjects.abstract.AbstractVideoDevice import AbstractVideoDevice


class QtTangoLimaVideo(AbstractVideoDevice):
    """
    Descript. :
    """

    def __init__(self, name):
        """
        Descript. :
        """
        AbstractVideoDevice.__init__(self, name)
        self.device = None

    def init(self):
        """
        Descript. :
        """

        tangoname = self.get_property("tangoname")

        width = self.get_property("width")
        height = self.get_property("height")

        if None not in [width, height]:
            self.width = width
            self.height = height
        else:
            self.width = None
            self.height = None

        self.device = PyTango.DeviceProxy(tangoname)
        self.device.ping()

        AbstractVideoDevice.init(self)

    def set_cam_encoding(self, cam_encoding):
        if cam_encoding == "yuv422p":
            self.device.video_mode = "YUV422"
        elif cam_encoding == "y8":
            self.device.video_mode = "Y8"
        elif cam_encoding.lower() == "bayer_rg16":
            self.device.video_mode = "BAYER_RG16"
        AbstractVideoDevice.set_cam_encoding(self, cam_encoding)

    """ Overloading of AbstractVideoDevice methods """

    def get_raw_image_size(self):
        # in case width and height not set in xml return server values
        if None not in [self.width, self.height]:
            return [self.width, self.height]
        else:
            return [self.device.image_width, self.device.image_height]

    def get_image(self):
        img_data = self.device.video_last_image

        if img_data[0] == "VIDEO_IMAGE":
            header_fmt = ">IHHqiiHHHH"
            _, ver, img_mode, frame_number, width, height, _, _, _, _ = struct.unpack(
                header_fmt, img_data[1][: struct.calcsize(header_fmt)]
            )
            raw_buffer = np.fromstring(img_data[1][32:], np.uint16)
        return raw_buffer, width, height

    def get_gain(self):
        if self.get_cam_type() == "basler":
            value = self.device.video_gain
            return value

    def set_gain(self, gain_value):
        if self.get_cam_type() == "basler":
            self.device.video_gain = gain_value
            return

    def get_exposure_time(self):
        if self.get_cam_type() == "basler":
            return self.device.video_exposure

    def set_exposure_time(self, exposure_time_value):
        if self.get_cam_type() == "basler":
            self.device.video_exposure = exposure_time_value

    def get_video_live(self):
        return self.device.video_live

    def set_video_live(self, flag):
        self.device.video_live = flag

    """ END Overloading of AbstractVideoDevice methods """

    def get_width(self):
        if self.width:
            return self.width
        else:
            return self.device.width

    def get_height(self):
        if self.height:
            return self.height
        else:
            return self.device.height


def test_hwo(hwo):
    print(("Image dimensions: %s" % hwo.get_image_dimensions()))
    print(("Live Mode: %s" % hwo.get_video_live()))



#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#   You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

import time

import numpy as np
from pkg_resources import resource_filename

from mxcubecore.HardwareObjects.abstract.AbstractVideoDevice import AbstractVideoDevice
from mxcubecore.utils.qt_import import (
    QBrush,
    QImage,
    QPainter,
    QPen,
    QPixmap,
    Qt,
)


class QtVideoMockup(AbstractVideoDevice):
    def __init__(self, name):
        AbstractVideoDevice.__init__(self, name)
        self.force_update = None
        self.image_type = None
        self.image = None

    def init(self):
        default_image_path = resource_filename(
            "mxcubecore", "configuration/mockup/qt/fakeimg.jpg"
        )
        image_path = self.get_property("file_name", default_image_path)
        self.image = QPixmap(image_path)
        self.image_dimensions = (self.image.width(), self.image.height())
        self.painter = QPainter(self.image)

        custom_pen = QPen(Qt.SolidLine)
        custom_pen.setColor(Qt.black)
        custom_pen.setWidth(1)
        self.painter.setPen(custom_pen)

        custom_brush = QBrush(Qt.SolidPattern)
        custom_brush.setColor(Qt.lightGray)
        self.painter.setBrush(custom_brush)

        AbstractVideoDevice.init(self)

    def get_new_image(self):
        self.painter.drawRect(self.image.width() - 75, self.image.height() - 30, 70, 20)
        self.painter.drawText(
            self.image.width() - 70, self.image.height() - 15, time.strftime("%H:%M:%S")
        )
        self.emit("imageReceived", self.image)

    def save_snapshot(self, filename, image_type="PNG"):
        qimage = QImage(self.image)
        qimage.save(filename, image_type)

    def get_snapshot(self, bw=None, return_as_array=None):
        qimage = QImage(self.image)
        if return_as_array:
            qimage = qimage.convertToFormat(4)
            ptr = qimage.bits()
            ptr.setsize(qimage.byteCount())

            image_array = np.array(ptr).reshape(qimage.height(), qimage.width(), 4)
            if bw:
                return np.dot(image_array[..., :3], [0.299, 0.587, 0.144])
            else:
                return image_array
        else:
            if bw:
                return qimage.convertToFormat(QImage.Format_Mono)
            else:
                return qimage

    def get_contrast(self):
        return 34

    def set_contrast(self, contrast_value):
        return

    def get_brightness(self):
        return 54

    def set_brightness(self, brightness_value):
        return

    def get_gain(self):
        return 32

    def set_gain(self, gain_value):
        return

    def get_gamma(self):
        return 22

    def set_gamma(self, gamma_value):
        return

    def get_exposure_time(self):
        return 0.23

    def get_video_live(self):
        return True

    def get_raw_image_size(self):
        return list(self.image_dimensions)



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU General Lesser Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

"""Abstract XRF spectrum class. Compliant with queue_entry/xrf_spectrum.py"""

import abc
import logging
import time
from pathlib import Path

import gevent

from mxcubecore import HardwareRepository as HWR
from mxcubecore.BaseHardwareObjects import HardwareObject

__copyright__ = """ Copyright © by the MXCuBE collaboration """
__license__ = "LGPLv3+"


class AbstractXRFSpectrum(HardwareObject):
    """Abstract XRFSpectrum procedure.

    Emits:
        stateChanged: ("stateChanged", (state))
        xrfSpectrumStatusChanged: ("xrfSpectrumStatusChanged", (error_msg)

    Attributes:
        default_integration_time (float): Time [s]
        spectrum_info_dict (dict): keys defined by the lims model.
        lims: reference to the lims hardware object

    States:
        HardwareObjectStates: READY, BUSY, FAULT

    Note:
        _execute_spectrum and spectrum_analyse are hooks to be overloaded
        for specific implementation.
    """

    __metaclass__ = abc.ABCMeta

    def __init__(self, name):
        super().__init__(name)
        self.lims = None
        self.spectrum_info_dict = {}
        self.default_integration_time = None
        self.cpos = None

    def init(self):
        """Initialisation"""
        self.default_integration_time = self.get_property("default_integration_time", 3)
        self.file_suffix = self.get_property("file_suffix", "dat")
        self.lims = HWR.beamline.lims
        if not self.lims:
            logging.getLogger().warning("XRFSpectrum: no lims set")

    def start_spectrum(
        self,
        integration_time: float | None = None,
        data_dir: str | None = None,
        prefix: str | None = None,
        archive_dir: str | None = None,
        session_id: int | None = None,
        blsample_id: int | None = None,
        cpos: dict | None = None,
    ):
        """Start the procedure. Called by the queue_model.

        Args:
            integration_time: Inregration time [s].
            data_dir: Directory to save the data (full path).
            archive_dir: Directory to save the archive data (full path).
            prefix: File prefix
            session_id: Session ID number (from ISpyB)
            blsample_id: Sample ID number (from ISpyB)
            cpos: The centred position motors and their values.
        """
        self.cpos = cpos
        self.spectrum_info_dict = {"sessionId": session_id, "blSampleId": blsample_id}
        integration_time = integration_time or self.default_integration_time
        self.spectrum_info_dict["exposureTime"] = integration_time
        self.spectrum_info_dict["filename"] = ""

        # Create the data and the archive directory (if needed) and files
        if data_dir:
            if not self.create_directory(data_dir):
                self.update_state(self.STATES.FAULT)
                return False
            filename = self.get_filename(data_dir, prefix)
            self.spectrum_info_dict["filename"] = filename + "." + self.file_suffix
        if archive_dir:
            if not self.create_directory(archive_dir):
                self.update_state(self.STATES.FAULT)
                return False
            filename = self.get_filename(archive_dir, prefix)
            self.spectrum_info_dict["scanFileFullPath"] = (
                filename + "." + self.file_suffix
            )
            self.spectrum_info_dict["jpegScanFileFullPath"] = filename + ".png"
            self.spectrum_info_dict["annotatedPymcaXfeSpectrum"] = filename + ".html"
            self.spectrum_info_dict["fittedDataFileFullPath"] = filename + "_peaks.csv"

        self.spectrum_info_dict["startTime"] = time.strftime("%Y-%m-%d %H:%M:%S")
        self.update_state(self.STATES.BUSY)

        gevent.spawn(
            self.execute_spectrum,
            integration_time,
            self.spectrum_info_dict["filename"],
        )
        return True

    def execute_spectrum(
        self,
        integration_time: float | None = None,
        filename: str | None = None,
    ):
        """Do the acquisition.

        Args:
            integration_time: MCA integration time [s].
            filename: Data file (full path).
        Raises:
            RuntimeError: Cannot acquire data.
        """
        if filename:
            self.spectrum_info_dict["filename"] = filename

        integration_time = integration_time or self.default_integration_time

        try:
            if self._execute_spectrum(integration_time, filename):
                self.spectrum_command_finished()
        except RuntimeError as err:
            msg = f"XRFSpectrum: could not acquire spectrum, {err}"
            logging.getLogger("user_level_log").exception(msg)
            self.spectrum_status_change(msg)
            self.update_state(self.STATES.FAULT)

    @abc.abstractmethod
    def _execute_spectrum(
        self,
        integration_time: float | None = None,
        filename: str | None = None,
    ) -> bool:
        """Specific XRF acquisition procedure"""
        return True

    def create_directory(self, directory: str) -> bool:
        """Create a directory, if needed.
        Args:
            directory: Directory to save the data (full path).
        Returns:
           ``True`` if directory created or already exists, ``False`` if error.
        """
        if not Path(directory).is_dir():
            msg = f"XRFSpectrum: directory creating {directory}"
            try:
                if not Path(directory).exists():
                    logging.getLogger("user_level_log").debug(msg)
                    Path(directory).mkdir(parents=True)
                return True
            except OSError as err:
                msg += f": {err}"
                logging.getLogger().error(msg)
                self.spectrum_status_change("Error creating directory")
                self.spectrum_command_aborted()
                return False
        return True

    def get_filename(self, directory: str, prefix: str) -> str:
        """Create file template.
        Args:
            directory(str): directory name (full path)
        Returns:
            (str): File template
        """
        _pattern = f"{prefix}_{time.strftime('%d_%b_%Y')}_%02d_xrf"
        filename = Path(directory) / (_pattern % 1)

        i = 2
        while Path(filename).is_file():
            filename = Path(directory) / (_pattern % i)
            i += 1

        return str(filename)

    def spectrum_status_change(self, status_msg: str):
        """Emit the signal xrfSpectrumStatusChanged with appropriate message.
        Args:
            status_msg(str): Message to send.
        """
        self.emit("xrfSpectrumStatusChanged", (status_msg,))

    def spectrum_command_finished(self):
        """Actions to do if spectrum acquired."""
        self.spectrum_info_dict["endTime"] = time.strftime("%Y-%m-%d %H:%M:%S")
        if HWR.beamline.transmission:
            self.spectrum_info_dict["beamTransmission"] = (
                HWR.beamline.transmission.get_value()
            )
        if HWR.beamline.energy:
            self.spectrum_info_dict["energy"] = HWR.beamline.energy.get_value()
        if HWR.beamline.flux:
            self.spectrum_info_dict["flux"] = HWR.beamline.flux.get_value()
        if HWR.beamline.beam:
            size = HWR.beamline.beam.get_value()
            self.spectrum_info_dict["beamSizeHorizontal"] = size[0]
            self.spectrum_info_dict["beamSizeVertical"] = size[1]
        self.spectrum_analyse()
        if self.lims:
            self.spectrum_store_lims()
        self.update_state(self.STATES.READY)

    def spectrum_analyse(self):
        """Get the spectrum data. Do analysis and save fitted data.
        The method has to be implemented as specific for each site, but
        is only optional.
        """

    def spectrum_command_aborted(self):
        """Spectrum aborted actions"""
        self.update_state(self.STATES.READY)

    def spectrum_command_failed(self):
        """Spectrum failed actions"""
        self.spectrum_info_dict["endTime"] = time.strftime("%Y-%m-%d %H:%M:%S")
        if self.lims:
            self.spectrum_store_lims()
        self.update_state(self.STATES.FAULT)

    def spectrum_store_lims(self):
        """Store the data in lims, according to the existing data model."""
        if self.spectrum_info_dict.get("sessionId"):
            self.lims.store_xfe_spectrum(self.spectrum_info_dict)



"""MAXIV Session hardware object."""

import time
from pathlib import Path

from sdm import storage

import mxcubecore.HardwareObjects.Session
from mxcubecore.model.queue_model_objects import PathTemplate


class Session(mxcubecore.HardwareObjects.Session.Session):
    def init(self):
        super().init()
        self.is_commissioning = False

        # We initialize with an empty archive_folder,
        # to be set once proposal is selected.
        PathTemplate.set_archive_path(self.base_archive_directory, "")

    def set_in_commissioning(self, proposal_info):
        self.proposal_code = proposal_info["Proposal"]["code"]
        self.proposal_number = proposal_info["Proposal"]["number"]
        self.is_commissioning = True

    def get_proposal(self) -> str:
        """
        Returns:
            The proposal, 'local-user' if no proposal is available
        """
        proposal = "local-user"

        if self.proposal_code and self.proposal_number:
            if self.proposal_code == "ifx":
                self.proposal_code = "fx"

            proposal = str(self.proposal_number)

        return proposal

    def get_base_data_directory(self) -> str:
        """Get base data directory.

        Figure out the base data directory taking the 'contextual'
        information into account, such as if the current user
        is in-house.

        Returns:
            The base data path.
        """
        # /data/(user-type)/(beamline)/(proposal)/(visit)/raw
        if self.session_start_date:
            start_time = self.session_start_date.split(" ")[0].replace("-", "")
        else:
            start_time = time.strftime("%Y%m%d")
        _proposal = self.get_proposal()

        if not self.is_commissioning:
            if self.is_proprietary():
                directory = Path(
                    self.base_directory,
                    "proprietary",
                    self.beamline_name.lower(),
                    _proposal,
                    start_time,
                )
            else:
                directory = Path(
                    self.base_directory,
                    "visitors",
                    self.beamline_name.lower(),
                    _proposal,
                    start_time,
                )
        else:
            # /data/staff/biomax/commissioning/date
            directory = Path(
                self.base_directory,
                "staff",
                self.beamline_name.lower(),
                "commissioning",
                time.strftime("%Y%m%d"),
            )

        self.log.info(
            "Data directory for proposal %s: %s",
            self.get_proposal(),
            directory,
        )

        return str(directory)

    def prepare_directories(self, session):
        self.log.info("Preparing Data directory for session: %s", session)
        start_date = session.start_datetime.date().isoformat().replace("-", "")
        self.set_session_start_date(start_date)

        self.log.info("Preparing Data directory for proposal %s", self.get_proposal())
        if self.is_commissioning:
            category = "staff"
        elif self.is_proprietary():
            category = "proprietary"
        else:
            category = "visitors"

        try:
            self.storage = storage.Storage(
                user_type=category, beamline=self.endstation_name
            )
        except Exception:
            self.log.exception("error setting up SDM")

        # This creates the path for the data and ensures proper permissions
        # e.g. /data/visitors/biomax/<proposal>/<visit>/{raw, process}
        if self.is_commissioning:
            group = self.beamline_name.lower()
        else:
            group = self.storage.get_proposal_group(session.number)
        try:
            _raw_path = self.storage.create_path(
                session.number, group, self.get_session_start_date()
            )

            self.log.info("SDM Data directory created: %s", _raw_path)
        except Exception as exc:  # noqa: BLE001
            self.log.warning("SDM Data directory creation failed. %s", exc)
            self.log.info("SDM Data directory trying to create again after failure")
            time.sleep(0.1)
            try:
                _raw_path = self.storage.create_path(
                    session.number, group, self.get_session_start_date()
                )

                self.log.info("SDM Data directory created: %s", _raw_path)
            except Exception:
                self.log.exception("SDM Data directory creation failed.")
                raise

        if self.base_archive_directory:
            archive_folder = "{}/{}".format(category, self.beamline_name.lower())
            PathTemplate.set_archive_path(self.base_archive_directory, archive_folder)
            _archive_path = Path(self.base_archive_directory, archive_folder)
            self.log.info("Archive directory configured: %s", _archive_path)

    def is_proprietary(self) -> bool:
        """Determines if current proposal is considered to be proprietary.

        Returns:
            True if the proposal is proprietary, otherwise False.
        """
        return self.proposal_code == "IN"

    def clear_session(self):
        self.session_id = None
        self.proposal_code = None
        self.proposal_number = None
        self.proposal_id = None
        self.is_commissioning = False



# pylint: skip-file
# ruff: noqa: F401  # This file is full of "unused imports" by design

"""

Description
--------------

This module import all symbols from PyQt in one single module
space

It also imports the *matplotlib* module and allow to check
version compatibility between PyQt and matplotbil

This module, on importing,  assigns the variables:

   qt_imported = [True,False]
   qt_version_no = <list of integers> example:  [4,8,1]
   qt_variant = ['PyQt5','PyQt4','PySide']

   mpl_imported = [True,False]
   mpl_version = <String as provided by matplotlib module>
   mpl_version_no = <list of integers> example:  [1,4,0]

Usage
-------------
This module could be imported directly instead of importing
PyQt.  But the original submodules in PyQt will be flattened.

For example::

   from PyQt4.QtCore import *
   from PyQt4.QtGui import *

   from PyQt5.QtCore import *
   from PyQt5.QtWidgets import *
   from PyQt5.QtGui import *

could be all replaced by::

   from qt_import import *

Or you could directly import things in this way::

  import qt_import as Qt
  app = Qt.QApplication([])

But import lines like::
   from qt_import import Qtcore

would fail (submodules are flattened)

To check the modules you are importing you can
directly use this module to print out a report::

   user@host: python qt_import.py
   Using PyQt:  PyQt4, version: 4.8.6
   Matplotlib:  PyQt4, version: 1.3.1
   Matplotlib is COMPATIBLE with PyQt




Credits
-------------
Inspired from code from "splot" (http://certif.com)
splot is licensed under LGPL license

"""

import os
import sys

qt_imported = False
qt_variant = None
qt_version_no = []

mpl_imported = False
mpl_version = None
mpl_version_no = False
#
#  Matplotlib import
#
try:
    import matplotlib

    mpl_imported = True
    mpl_version = matplotlib.__version__
    version_parts = mpl_version.split(".")
    mpl_major, mpl_minor = version_parts[:2]
    mpl_version_no = [int(mpl_major), int(mpl_minor), 0]

    if len(version_parts) > 2:
        try:
            import re

            rel = version_parts[2]
            m = re.search(r"(?P<release>\d+)", rel)
            if m:
                mpl_version_no[2] = int(m.group("release"))
        except BaseException:
            pass
except BaseException:
    pass

if "--pyqt5" in sys.argv:
    qt_variant = "PyQt5"
elif "--pyside" in sys.argv:
    qt_variant = "PySide"
elif "--pyqt4" in sys.argv:
    qt_variant = "PyQt4"
elif "--pyqt3" in sys.argv:
    qt_variant = "PyQt3"

#
# PyQt5
#
if (qt_variant == "PyQt5") or (qt_variant is None and not qt_imported):
    try:
        from PyQt5.QtCore import (
            PYQT_VERSION_STR,
            QT_VERSION_STR,
            QCoreApplication,
            QDir,
            QEvent,
            QEventLoop,
            QObject,
            QPoint,
            QPointF,
            QRect,
            QRectF,
            QRegExp,
            QSize,
            Qt,
            QTimer,
            pyqtSignal,
            pyqtSlot,
        )
        from PyQt5.QtGui import (
            QBrush,
            QColor,
            QContextMenuEvent,
            QCursor,
            QDoubleValidator,
            QFocusEvent,
            QFont,
            QIcon,
            QImage,
            QIntValidator,
            QKeyEvent,
            QKeySequence,
            QLinearGradient,
            QMouseEvent,
            QPainter,
            QPainterPath,
            QPalette,
            QPen,
            QPixmap,
            QPolygon,
            QRegExpValidator,
            QValidator,
        )
        from PyQt5.QtWidgets import (
            QAbstractItemView,
            QAction,
            QActionGroup,
            QApplication,
            QButtonGroup,
            QCheckBox,
            QColorDialog,
            QComboBox,
            QDesktopWidget,
            QDial,
            QDialog,
            QDoubleSpinBox,
            QFileDialog,
            QFrame,
            QGraphicsItem,
            QGraphicsPixmapItem,
            QGraphicsScene,
            QGraphicsView,
            QGridLayout,
            QGroupBox,
            QHBoxLayout,
            QHeaderView,
            QInputDialog,
            QLabel,
            QLayout,
            QLineEdit,
            QListView,
            QListWidget,
            QListWidgetItem,
            QMainWindow,
            QMenu,
            QMenuBar,
            QMessageBox,
            QProgressBar,
            QProgressDialog,
            QPushButton,
            QRadioButton,
            QScrollArea,
            QScrollBar,
            QSizePolicy,
            QSlider,
            QSpacerItem,
            QSpinBox,
            QSplashScreen,
            QSplitter,
            QStackedWidget,
            QStatusBar,
            QTableView,
            QTableWidget,
            QTableWidgetItem,
            QTabWidget,
            QTextBrowser,
            QTextEdit,
            QToolBar,
            QToolBox,
            QToolButton,
            QToolTip,
            QTreeView,
            QTreeWidget,
            QTreeWidgetItem,
            QTreeWidgetItemIterator,
            QVBoxLayout,
            QWhatsThis,
            QWidget,
        )
        from PyQt5.uic import loadUi

        QStringList = list
        getQApp = QCoreApplication.instance
        qApp = QCoreApplication.instance

        qt_imported = True
        qt_variant = "PyQt5"
        qt_version_no = list(map(int, QT_VERSION_STR.split(".")))
        _ver = PYQT_VERSION_STR.split(".")
        ver = _ver + ["0"] * (3 - len(_ver))
        pyqt_version_no = list(map(int, ver))[:3]
    except ImportError:
        pass

    try:
        from PyQt5.QtWebKit import QWebPage
    except ImportError:
        pass

#
# PyQt4
#
if (qt_variant == "PyQt4") or (qt_variant is None and not qt_imported):
    #
    # Maybe it is PyQt4 that you want
    #

    # force api 2 (so that the api is common for all pyqt versions)
    #   !! this means the classes below will not exist !!
    # but code is guaranteed to be compatible
    try:
        from PyQt4.QtCore import (
            PYQT_VERSION_STR,
            QT_VERSION_STR,
            SIGNAL,
            QDir,
            QEvent,
            QEventLoop,
            QObject,
            QPoint,
            QPointF,
            QRect,
            QRectF,
            QRegExp,
            QSize,
            QStringList,
            Qt,
            QTimer,
            QUrl,
            pyqtSignal,
            pyqtSlot,
        )
        from PyQt4.QtGui import (
            QAbstractItemView,
            QAction,
            QActionGroup,
            QApplication,
            QBrush,
            QButtonGroup,
            QCheckBox,
            QColor,
            QColorDialog,
            QComboBox,
            QContextMenuEvent,
            QCursor,
            QDesktopWidget,
            QDial,
            QDialog,
            QDoubleSpinBox,
            QDoubleValidator,
            QFileDialog,
            QFocusEvent,
            QFont,
            QFrame,
            QGraphicsItem,
            QGraphicsPixmapItem,
            QGraphicsScene,
            QGraphicsView,
            QGridLayout,
            QGroupBox,
            QHBoxLayout,
            QHeaderView,
            QIcon,
            QImage,
            QInputDialog,
            QIntValidator,
            QKeyEvent,
            QKeySequence,
            QLabel,
            QLayout,
            QLinearGradient,
            QLineEdit,
            QListView,
            QListWidget,
            QListWidgetItem,
            QMainWindow,
            QMenu,
            QMenuBar,
            QMessageBox,
            QMouseEvent,
            QPainter,
            QPainterPath,
            QPalette,
            QPen,
            QPixmap,
            QPolygon,
            QProgressBar,
            QProgressDialog,
            QPushButton,
            QRadioButton,
            QRegExpValidator,
            QScrollArea,
            QScrollBar,
            QSizePolicy,
            QSlider,
            QSpacerItem,
            QSpinBox,
            QSplashScreen,
            QSplitter,
            QStackedWidget,
            QStatusBar,
            QTableView,
            QTableWidget,
            QTableWidgetItem,
            QTabWidget,
            QTextBrowser,
            QTextEdit,
            QToolBar,
            QToolBox,
            QToolButton,
            QToolTip,
            QTreeView,
            QTreeWidget,
            QTreeWidgetItem,
            QTreeWidgetItemIterator,
            QValidator,
            QVBoxLayout,
            QWhatsThis,
            QWidget,
            qApp,
        )
        from PyQt4.uic import loadUi

        def getQApp():
            return qApp

        qt_imported = True
        qt_variant = "PyQt4"
        qt_version_no = list(map(int, QT_VERSION_STR.split(".")))
        _ver = PYQT_VERSION_STR.split(".")
        ver = _ver + ["0"] * (3 - len(_ver))
        pyqt_version_no = list(map(int, ver))[:3]
    except BaseException:
        pass

    try:
        from PyQt4.QtWebKit import QWebPage
    except ImportError:
        pass

#
# PySide
#
if (qt_variant == "PySide") or (qt_variant is None and not qt_imported):
    #
    # Finally try PySide (not fully tested)
    #
    try:
        from PySide import __version_info__

        pyqt_version_no = __version_info__[:3]
        from PySide.QtCore import __version_info__

        qt_version_no = __version_info__[:3]

        from PySide.QtCore import *
        from PySide.QtGui import *
        from PySide.QtSvg import *
        from PySide.QtUiTools import *
        from PySide.QtWebKit import *

        pyqtSignal = Signal
        pyqtSlot = Slot
        qt_imported = True
        qt_variant = "PySide"
        getQApp = QCoreApplication.instance

        def loadUi(filename, parent=None):
            loader = QUiLoader()
            uifile = QFile(filename)
            uifile.open(QFile.ReadOnly)
            ui = loader.load(uifile, parent)
            uifile.close()
            return ui

    except BaseException:
        pass

#
#  Matplotlib backend assignment
#
if mpl_imported:
    if qt_variant == "PyQt5":
        if mpl_version_no < [1, 4, 0]:
            mpl_compat = False
        else:
            mpl_compat = True
            matplotlib.use("Qt5Agg")

    elif qt_variant == "PySide":
        if mpl_version_no < [1, 1, 0]:
            mpl_compat = False
        else:
            mpl_compat = True
            matplotlib.use("Qt4Agg")
            from matplotlib import rcParams

            rcParams["backend.qt4"] = "PySide"

    elif qt_variant == "PyQt4":
        mpl_compat = True
        matplotlib.use("Qt4Agg")

if "QString" not in globals():
    QString = str

"""
if qt_imported:
    print("Using PyQt:  %s" % (qt_variant))
    print("   qt version: %s / pyqt version: %s" % ("%d.%d.%d" % tuple(qt_version_no), "%d.%d.%d" % tuple(pyqt_version_no)))
else:
    print("Cannot import PyQt")

if mpl_imported:
    print("Matplotlib version: %s" % "%d.%d.%d" % tuple(mpl_version_no))
else:
    print("Cannot import matplotlib")

if qt_imported and mpl_imported:
    if mpl_compat:
        print("  Matplotlib is COMPATIBLE with PyQt")
    else:
        print("  !!! Matplotlib is NOT COMPATIBLE with PyQt !!!")
"""

if qt_variant in ("PyQt4", "PyQt5", "PySide"):
    # QHeaderView is not defined for Qt3, so the import broke without the 'if'
    class RotatedHeaderView(QHeaderView):
        def __init__(self, parent=None):
            super(RotatedHeaderView, self).__init__(Qt.Horizontal, parent)
            self.setMinimumSectionSize(22)

        def paintSection(self, painter, rect, logicalIndex):
            painter.save()
            # translate the painter such that rotate will rotate around the correct
            # point
            painter.translate(rect.x() + rect.width(), rect.y())
            painter.rotate(90)
            # and have parent code paint at this location
            newrect = QRect(0, 0, rect.height(), rect.width())
            super(RotatedHeaderView, self).paintSection(painter, newrect, logicalIndex)
            painter.restore()

        def minimumSizeHint(self):
            size = super(RotatedHeaderView, self).minimumSizeHint()
            size.transpose()
            return size

        def sectionSizeFromContents(self, logicalIndex):
            size = super(RotatedHeaderView, self).sectionSizeFromContents(logicalIndex)
            size.transpose()
            return size

    class QDoubleSlider(QSlider):
        doubleValueChanged = pyqtSignal(float)

        def __init__(self, orientation=Qt.Horizontal, parent=None):
            super(QSlider, self).__init__(orientation, parent)
            self.decimals = 5
            self._max_int = 10**self.decimals

            super(QSlider, self).setMinimum(0)
            super(QSlider, self).setMaximum(self._max_int)

            self._min_value = 0.0
            self._max_value = 1.0

            self.valueChanged.connect(self.value_changed)

        def value_changed(self, value):
            self.doubleValueChanged.emit(value / float(self._max_int))

        @property
        def _value_range(self):
            return self._max_value - self._min_value

        def value(self):
            return (
                float(super(QSlider, self).value()) / self._max_int * self._value_range
                + self._min_value
            )

        def setValue(self, value):
            super(QSlider, self).setValue(
                int((value - self._min_value) / self._value_range * self._max_int)
            )

        def setMinimum(self, value):
            if value > self._max_value:
                raise ValueError("Minimum limit cannot be higher than maximum")

            self._min_value = value
            # self.setValue(self.value())

        def setMaximum(self, value):
            if value < self._min_value:
                raise ValueError("Minimum limit cannot be higher than maximum")

            self._max_value = value
            # self.setValue(self.value())

        def minimum(self):
            return self._min_value

        def maximum(self):
            return self._max_value


def load_ui_file(filename):
    current_path = os.path.dirname(os.path.abspath(__file__)).split(os.sep)
    current_path = os.path.join(*current_path[1:-1])
    return loadUi(os.path.join("/", current_path, "ui_files", filename))



from __future__ import print_function

import itertools
import uuid

from mxcubecore.HardwareObjects.abstract.ISPyBAbstractLims import ISPyBAbstractLIMS
from mxcubecore.model.lims_session import (
    LimsSessionManager,
    Proposal,
    Session,
)

"""
A client for ISPyB Webservices.
"""

import logging


class ProposalTypeISPyBLims(ISPyBAbstractLIMS):
    """
    ISPyB proposal-based client
    """

    def __init__(self, name):
        super().__init__(name)

    def is_user_login_type(self):
        return False

    def get_proposals_by_user(self, login_id: str):
        raise Exception("Not implemented")

    def get_full_user_name(self):
        return self.get_user_name()

    def get_user_name(self):
        """
        Because it is a proposal based it returns the proposal plus the uuid4
        """
        return f"{self.user_name}-{str(uuid.uuid4())}"

    def _authenticate(self, user_name, psd):
        """
        Authentication step based on the authServerType
        """

        if self.authServerType == "ldap":
            logging.getLogger("HWR").debug(
                "Starting LDAP authentication %s" % user_name
            )
            ok = self.ldap_login(user_name, psd)
        elif self.authServerType == "ispyb":
            logging.getLogger("HWR").debug("ISPyB login")
            ok, _ = self.ispyb_login(user_name, psd)
        else:
            raise Exception("Authentication server type is not defined")

        if not ok:
            # refuse Login
            logging.getLogger("HWR").error(
                "Authentication with %s failed" % self.authServerType
            )
            raise Exception("Authentication failed")

        logging.getLogger("HWR").debug(
            "User %s logged in %s" % (user_name, self.authServerType)
        )

    def is_session_already_active_by_code(self, code: str, number: str) -> bool:
        # If current selected session is already selected no need to do
        # anything else
        active_session = self.session_manager.active_session
        if active_session is not None:
            if (
                code.upper() == active_session.code.upper()
                and number.upper() == active_session.number.upper()
            ):
                return True
        return False

    def set_active_session_by_id(self, proposal_name: str) -> Session:
        """
        Given a proposal name it will select a session that is scheduled on this beamline in the current timeslot
        """
        if self.is_session_already_active(self.session_manager.active_session):
            logging.getLogger("HWR").debug("[ISPYB] is_session_already_active")
            return self.session_manager.active_session

        if len(self.session_manager.sessions) == 0:
            logging.getLogger("HWR").error(
                "Session list is empty. No session candidates"
            )
            raise Exception("No sessions available")

        if len(self.session_manager.sessions) == 1:
            self.session_manager.active_session = self.session_manager.sessions[0]
            logging.getLogger("HWR").debug(
                "Session list contains a single session. proposal_name=%s",
                self.session_manager.active_session.proposal_name,
            )
            return self.session_manager.active_session

        session_list = [
            obj
            for obj in self.session_manager.sessions
            if (
                obj.proposal_name.upper() == proposal_name.upper()
                and obj.is_scheduled_beamline
                and obj.is_scheduled_time
            )
        ]
        if len(session_list) > 1:
            logging.getLogger("HWR").warning(
                "Session not found in the local list of sessions. Found %s sessions. proposal_name=%s"
                % (len(session_list), proposal_name)
            )
        self.session_manager.active_session = session_list[0]
        return self.session_manager.active_session

    def _get_proposal_code_and_number_by_proposal_name(self, proposal_name):
        code = "".join(itertools.takewhile(lambda c: not c.isdigit(), proposal_name))
        number = proposal_name[len(code) :]
        return [code, number]

    def get_session_manager_by_code_number(
        self, code: str, number: str, is_local_host: bool
    ) -> LimsSessionManager:
        logging.getLogger("HRW").debug("[ISPyB] get_session_manager_by_code_number")

        self.session_manager = self.adapter.get_sessions_by_code_and_number(
            code, number, self.beamline_name
        )
        proposal_name = code + number
        # If there is no session then a session is created
        if len(self.session_manager.sessions) == 0 and is_local_host:
            logging.getLogger("HRW").debug(
                "No sessions found for proposal=%s" % proposal_name
            )
            logging.getLogger("HRW").debug(
                "Creating session for proposal=%s" % proposal_name
            )
            proposal: Proposal = self.adapter.find_proposal(code, number)
            self.create_session(proposal.proposal_id)
            self.session_manager = self.adapter.get_sessions_by_code_and_number(
                code, number, self.beamline_name
            )
            logging.getLogger("HRW").debug(
                "Sessions count=%s" % len(self.session_manager.sessions)
            )

        self.set_active_session_by_id(proposal_name)

        return self.session_manager

    def get_session_manager_by_proposal_name(
        self, proposal_name: str, is_local_host: bool
    ) -> LimsSessionManager:
        code_number = self._get_proposal_code_and_number_by_proposal_name(proposal_name)
        code = self._translate(
            code_number[0],
            "ispyb",
        )
        number = code_number[1]

        return self.get_session_manager_by_code_number(code, number, is_local_host)

    def login(
        self, user_name: str, password: str, is_local_host: bool
    ) -> LimsSessionManager:
        logging.getLogger("HRW").debug(
            "Login on ISPyBLims proposal=%s is_local_host=%s"
            % (user_name, str(is_local_host)),
        )
        self.session_manager = LimsSessionManager()
        # For proposal login, split the loginID to code and numbers
        code = self._get_proposal_code_and_number_by_proposal_name(user_name)[0]
        number = self._get_proposal_code_and_number_by_proposal_name(user_name)[1]

        # if translation of the loginID is needed, need to be tested by ESRF
        if self.loginTranslate is True:
            user_name = self._translate(code, "ldap") + str(number)

        # Authentication
        try:
            self._authenticate(user_name, password)
            self.user_name = user_name
        except Exception as e:
            raise e

        return self.get_session_manager_by_proposal_name(user_name, is_local_host)



"""
Provides an API to add Command and Channel objects to hardware objects,
as specified in it's YAML configuration file.

See setup_commands_channels() function for details.
"""

from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Callable,
    Iterable,
)

if TYPE_CHECKING:
    from mxcubecore.BaseHardwareObjects import HardwareObject


def _setup_tango_commands_channels(hwobj: HardwareObject, tango_config: dict):
    """Set up Tango Command and Channel objects.

    parameters:
        tango: the 'tango' section of the hardware object's configuration
    """
    from mxcubecore.model.protocols.tango import (
        Device,
        TangoConfig,
    )

    def setup_tango_device(device_name: str, device_config: Device):
        #
        # set-up commands
        #
        for command_name, command_config in device_config.get_commands():
            attrs = {"type": "tango", "name": command_name, "tangoname": device_name}
            hwobj.add_command(attrs, command_config.name)

        #
        # set-up channels
        #
        for channel_name, channel_config in device_config.get_channels():
            attrs = {"type": "tango", "name": channel_name, "tangoname": device_name}

            if channel_config.polling_period:
                attrs["polling"] = channel_config.polling_period

            if channel_config.timeout:
                attrs["timeout"] = channel_config.timeout

            hwobj.add_channel(attrs, channel_config.attribute)

    tango_cfg = TangoConfig.model_validate(tango_config)
    for device_name, device_config in tango_cfg.get_tango_devices():
        setup_tango_device(device_name, device_config)


def _setup_exporter_commands_channels(hwobj: HardwareObject, exporter_config: dict):
    from mxcubecore.model.protocols.exporter import (
        Address,
        ExporterConfig,
    )

    def setup_address(address: str, address_config: Address):
        #
        # set-up commands
        #
        for command_name, command_config in address_config.get_commands():
            attrs = {
                "type": "exporter",
                "exporter_address": address,
                "name": command_name,
            }
            hwobj.add_command(attrs, command_config.name)

        #
        # set-up channels
        #
        for channel_name, channel_config in address_config.get_channels():
            attrs = {
                "type": "exporter",
                "exporter_address": address,
                "name": channel_name,
            }
            hwobj.add_channel(attrs, channel_config.attribute)

    exp_cfg = ExporterConfig.model_validate(exporter_config)
    for address, address_config in exp_cfg.get_addresses():
        setup_address(address, address_config)


def _setup_epics_channels(hwobj: HardwareObject, epics_config: dict):
    from mxcubecore.model.protocols.epics import (
        EpicsConfig,
        Prefix,
    )

    def setup_prefix(prefix: str, prefix_config: Prefix):
        #
        # set-up channels
        #
        for channel_name, channel_config in prefix_config.get_channels():
            attrs = {"type": "epics", "name": channel_name}
            if channel_config.polling_period:
                attrs["polling"] = channel_config.polling_period

            pv_name = f"{prefix}{channel_config.suffix}"
            hwobj.add_channel(attrs, pv_name)

    epics_cfg = EpicsConfig.model_validate(epics_config)
    for prefix, prefix_config in epics_cfg.get_prefixes():
        setup_prefix(prefix, prefix_config)


def _setup_sardana_commands(hwobj: HardwareObject, sardana_config: dict):
    from mxcubecore.model.protocols.sardana import (
        Door,
        SardanaConfig,
    )

    def setup_door(door_name: str, door_config: Door):
        for command_name, command_config in door_config.get_commands():
            attrs = {
                "type": "sardana",
                "doorname": door_name,
                "name": command_name,
            }
            hwobj.add_command(attrs, command_config.name)

    sardana_cfg = SardanaConfig.model_validate(sardana_config)
    for door_name, door_config in sardana_cfg.get_doors():
        setup_door(door_name, door_config)


def _protocol_handles():
    return {
        "tango": _setup_tango_commands_channels,
        "exporter": _setup_exporter_commands_channels,
        "epics": _setup_epics_channels,
        "sardana": _setup_sardana_commands,
    }


def _get_protocol_names() -> Iterable[str]:
    """Get names of all supported protocols."""
    return _protocol_handles().keys()


def _get_protocol_handler(protocol_name: str) -> Callable:
    """Get the callable that will set up commands and channels for a specific
    protocol.
    """
    return _protocol_handles()[protocol_name]


def _setup_protocol(hwobj: HardwareObject, config: dict, protocol: str):
    """Add the Command and Channel objects configured in the specified protocol section.

    parameters:
        protocol: name of the protocol to handle
    """
    protocol_config = config.get(protocol)
    if protocol_config is None:
        # no configuration for this protocol
        return

    _get_protocol_handler(protocol)(hwobj, protocol_config)


def setup_commands_channels(hwobj: HardwareObject, config: dict):
    """Add the Command and Channel objects to a hardware object, as specified i
       the config.

    parameters:
        hwobj: hardware object where to add Command and Channel objects
        config: the complete hardware object configuration, i.e. parsed YAML file
                as dict
    """
    for protocol in _get_protocol_names():
        _setup_protocol(hwobj, config, protocol)



"""LimaEigerDetector Class
Lima Tango Device Server implementation of the Dectris Eiger2 Detector.
"""

import logging
import math
import os
import time

import gevent

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.abstract.AbstractDetector import AbstractDetector
from mxcubecore.model.queue_model_objects import PathTemplate
from mxcubecore.TaskUtils import task


class LimaEigerDetector(AbstractDetector):
    def __init__(self, name):
        AbstractDetector.__init__(self, name)
        self.binning_mode = 1

    def init(self):
        AbstractDetector.init(self)

        self.header = {}
        self._images_per_file = self.get_property("images_per_file", 100)

        lima_device = self.get_property("lima_device")
        eiger_device = self.get_property("eiger_device")

        for channel_name in (
            "acq_status",
            "acq_trigger_mode",
            "saving_mode",
            "acq_nb_frames",
            "acq_expo_time",
            "saving_directory",
            "saving_prefix",
            "saving_suffix",
            "saving_next_number",
            "saving_index_format",
            "saving_format",
            "saving_overwrite_policy",
            "last_image_saved",
            "saving_frame_per_file",
            "saving_managed_mode",
        ):
            self.add_channel(
                {"type": "tango", "name": channel_name, "tangoname": lima_device},
                channel_name,
            )

        for channel_name in ("photon_energy",):
            self.add_channel(
                {"type": "tango", "name": channel_name, "tangoname": eiger_device},
                channel_name,
            )

        self.add_command(
            {"type": "tango", "name": "prepare_acq", "tangoname": lima_device},
            "prepareAcq",
        )
        self.add_command(
            {"type": "tango", "name": "start_acq", "tangoname": lima_device}, "startAcq"
        )
        self.add_command(
            {"type": "tango", "name": "stop_acq", "tangoname": lima_device}, "stopAcq"
        )
        self.add_command(
            {"type": "tango", "name": "reset", "tangoname": lima_device}, "reset"
        )
        self.add_channel(
            {"type": "tango", "name": "saving_common_header", "tangoname": lima_device},
            "saving_common_header",
        )

        self.get_command_object("prepare_acq").init_device()
        self.get_command_object("prepare_acq").device.set_timeout_millis(5 * 60 * 1000)
        self.get_channel_object("photon_energy").init_device()
        self.update_state(self.STATES.READY)

    def has_shutterless(self):
        return True

    def wait_ready(self, timeout=30):
        acq_status_chan = self.get_channel_object("acq_status")

        if acq_status_chan.get_value() != "Ready":
            self.update_state(self.STATES.BUSY)

        with gevent.Timeout(timeout, RuntimeError("Detector not ready")):
            while acq_status_chan.get_value() != "Ready":
                time.sleep(1)

        self.update_state(self.STATES.READY)

    def last_image_saved(self):
        return self.get_channel_object("last_image_saved").get_value() + 1

    def get_deadtime(self):
        return float(self.get_property("deadtime"))

    def prepare_acquisition(
        self,
        take_dark,
        start,
        osc_range,
        exptime,
        npass,
        number_of_images,
        comment,
        mesh,
        mesh_num_lines,
    ):
        """
        diffractometer_positions = HWR.beamline.diffractometer.get_positions()
        self.start_angles = list()
        for i in range(number_of_images):
            self.start_angles.append("%0.4f deg." % (start + osc_range * i))
        self.header["file_comments"] = comment
        self.header["N_oscillations"] = number_of_images
        self.header["Oscillation_axis"] = "omega"
        self.header["Chi"] = "0.0000 deg."
        try:
            self.header["Phi"] = "%0.4f deg." % diffractometer_positions.get(
                "kappa_phi", -9999
            )
            self.header["Kappa"] = "%0.4f deg." % diffractometer_positions.get(
                "kappa", -9999
            )
        except Exception:
            self.header["Phi"] = "0.0000 deg."
            self.header["Kappa"] = "0.0000 deg."
        self.header["Alpha"] = "0.0000 deg."
        self.header["Polarization"] = HWR.beamline.collect.bl_config.polarisation
        self.header["Detector_2theta"] = "0.0000 deg."
        self.header["Angle_increment"] = "%0.4f deg." % osc_range
        self.header["Transmission"] = HWR.beamline.transmission.get_value()
        self.header["Flux"] = HWR.beamline.flux.get_value()
        self.header["Detector_Voffset"] = "0.0000 m"
        self.header["Energy_range"] = "(0, 0) eV"
        self.header["Trim_directory:"] = "(nil)"
        self.header["Flat_field:"] = "(nil)"
        self.header["Excluded_pixels:"] = " badpix_mask.tif"
        self.header["N_excluded_pixels:"] = "= 321"
        self.header["Threshold_setting"] = (
            "%d eV" % self.get_channel_object("photon_energy").get_value()
        )
        self.header["Count_cutoff"] = "1048500"
        self.header["Tau"] = "= 0 s"
        self.header["Exposure_period"] = "%f s" % (exptime + self.get_deadtime())
        self.header["Exposure_time"] = "%f s" % exptime
        """

        self.stop()
        self.wait_ready()

        beam_x, beam_y = self.get_beam_position()
        header_info = [
            "beam_center_x=%s" % (beam_x),
            "beam_center_y=%s" % (beam_y),
            "detector_distance=%s"
            % (HWR.beamline.detector.distance.get_value() / 1000.0),
            "omega_start=%0.4f" % start,
            "omega_increment=%0.4f" % osc_range,
            "wavelength=%s" % HWR.beamline.energy.get_wavelength(),
        ]
        # Either we set the wavelength or we set the energy_threshold.
        # Up to now both ways are possible:
        # self.set_energy_threshold(HWR.beamline.energy.get_value())  # noqa: ERA001
        # "wavelength=%s" % HWR.beamline.energy.get_wavelength()  # noqa: ERA001

        self.get_channel_object("saving_common_header").set_value(header_info)

        if mesh:
            """
            self.get_channel_object("acq_trigger_mode").set_value("EXTERNAL_TRIGGER_SEQUENCES")
            self.get_channel_object("acq_nb_sequences").set_value(mesh_num_lines)
            """
            self.get_channel_object("acq_trigger_mode").set_value(
                "EXTERNAL_TRIGGER_MULTI"
            )
        else:
            self.set_channel_value("acq_trigger_mode", "EXTERNAL_TRIGGER")

        self.get_channel_object("saving_frame_per_file").set_value(
            min(self._images_per_file, number_of_images)
        )

        # 'MANUAL', 'AUTO_FRAME', 'AUTO_SEQUENCE
        self.get_channel_object("saving_mode").set_value("AUTO_FRAME")
        logging.info("Acq. nb frames = %d", number_of_images)
        self.get_channel_object("acq_nb_frames").set_value(number_of_images)
        self.get_channel_object("acq_expo_time").set_value(exptime)
        # 'ABORT', 'OVERWRITE', 'APPEND'
        self.get_channel_object("saving_overwrite_policy").set_value("OVERWRITE")
        # 'SOFTWARE', 'HARDWARE'
        self.get_channel_object("saving_managed_mode").set_value("HARDWARE")

    def set_energy_threshold(self, energy):
        minE = self.get_property("minE")
        if energy < minE:
            energy = minE

        working_energy_chan = self.get_channel_object("photon_energy")
        working_energy = working_energy_chan.get_value() / 1000.0
        if math.fabs(working_energy - energy) > 0.1:
            egy = int(energy * 1000.0)
            working_energy_chan.set_value(egy)

    @task
    def set_detector_filenames(self, frame_number, start, filename):
        prefix, suffix = os.path.splitext(os.path.basename(filename))
        prefix = "_".join(prefix.split("_")[:-1]) + "_"
        dirname = os.path.dirname(filename)
        if dirname.startswith(os.path.sep):
            dirname = dirname[len(os.path.sep) :]

        saving_directory = os.path.join(self.get_property("buffer"), dirname)
        self.wait_ready()

        self.get_channel_object("saving_directory").set_value(saving_directory)
        self.get_channel_object("saving_prefix").set_value(
            prefix + "%01d" % frame_number
        )
        self.get_channel_object("saving_suffix").set_value(suffix)
        self.get_channel_object("saving_format").set_value("HDF5")

    def start_acquisition(self):
        self.wait_ready()
        logging.getLogger("user_level_log").info("Preparing acquisition")
        self.get_command_object("prepare_acq")()
        logging.getLogger("user_level_log").info("Detector ready, continuing")
        self.get_command_object("start_acq")()

    def stop_acquisition(self):
        self.update_state(self.STATES.BUSY)
        try:
            self.get_command_object("stop_acq")()
        except Exception:
            pass

        time.sleep(1)
        self.get_command_object("reset")()
        self.wait_ready()
        self.update_state(self.STATES.READY)

    def reset(self):
        self.stop_acquisition()
        return True

    @property
    def status(self):
        try:
            acq_status = self.get_channel_value("acq_status")
        except Exception:
            acq_status = "OFFLINE"

        status = {"acq_satus": acq_status.upper()}

        return status

    def _emit_status(self):
        self.emit("statusChanged", self.status)

    def recover_from_failure(self):
        pass

    def get_image_file_name(self, pt, suffix=None):
        pt.precision = 1
        template = "%s_%s_%%" + str(pt.precision) + "d_master.%s"

        if suffix:
            file_name = template % (pt.get_prefix(), pt.run_number, suffix)
        else:
            file_name = template % (pt.get_prefix(), pt.run_number, pt.suffix)
        if pt.compression:
            file_name = "%s.gz" % file_name

        return file_name

    def get_first_and_last_file(self, pt: PathTemplate) -> tuple:
        """
        Get complete path to first and last image

        Args:
          Path template parameter

        Returns:
        Tuple containing first and last image path (first, last)
        """
        start_num = int(math.ceil(pt.start_num / 100))
        end_num = int(math.ceil((pt.start_num + pt.num_files - 1) / 100))

        return (pt.get_image_path() % start_num, pt.get_image_path() % end_num)

    def get_actual_file_path(self, master_file_path: str, image_number: int) -> tuple:
        """
        Get file path to image with the given image number <image_number> and
        the master h5 file <master_file_path>

        Args:
            first_file_path: Path to master h5 file
            image_number: image number (absolute)

        Returns:
            A tuple [image file path, image_number (relative to file)]
        """
        result_data_path = "_".join(master_file_path.split("_")[0:-2])
        img_number = int(image_number) % self._images_per_file
        start_file_number = int(int(image_number) / self._images_per_file) + 1
        result_data_path += "_data_%06d.h5" % start_file_number

        return result_data_path, img_number



# encoding: utf-8
#
#  Project name: MXCuBE
#  https://github.com/mxcube
#
#  This file is part of MXCuBE software.
#
#  MXCuBE is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  MXCuBE is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with MXCuBE. If not, see <http://www.gnu.org/licenses/>.

__copyright__ = """Copyright The MXCuBE Collaboration"""
__license__ = "LGPLv3+"

import logging
import time

from mxcubecore.HardwareObjects.abstract.AbstractTransmission import (
    AbstractTransmission,
)

log = logging.getLogger("HWR")


class P11Transmission(AbstractTransmission):
    """
    P11Transmission class for controlling transmission settings.

    This class extends the AbstractTransmission class and provides
    methods to interact with hardware for setting and reading the
    transmission values, as well as handling state changes.

    Attributes:
        chan_read_value (Channel): Channel object to read the current value.
        chan_set_value (Channel): Channel object to set the transmission value.
        chan_state (Channel): Channel object to track the state of the hardware.
    """

    def __init__(self, name):
        """
        Initializes the P11Transmission object with the given name.

        Args:
            name (str): The name of the transmission component.
        """
        super().__init__(name)
        self.chan_read_value = None
        self.chan_set_value = None
        self.chan_state = None

    def init(self):
        """
        Initializes the hardware channels and connects signals.

        This method sets up the communication channels for reading,
        setting, and tracking the transmission state. It connects
        signals to handle updates from the hardware.
        """
        self.chan_read_value = self.get_channel_object("chanRead")
        self.chan_set_value = self.get_channel_object("chanSet")
        self.chan_state = self.get_channel_object("chanState")

        if self.chan_read_value is not None:
            self.chan_read_value.connect_signal("update", self.re_emit_value)

        if self.chan_state is not None:
            self.chan_state.connect_signal("update", self.state_changed)

        self.re_emit_values()

    def re_emit_value(self, *args):
        """
        Re-emits the current transmission value and state.

        This method triggers the state and value updates to ensure
        the current hardware state and value are reflected in the software.

        Args:
            *args: Optional argument to handle extra signal data.
        """
        self.state_changed()
        # Value update is now handled in AbstractActuator, no need for value_changed method here

    def get_state(self):
        """
        Gets the current state of the transmission.

        Returns:
            str: The current state of the transmission ("READY", "BUSY", or "FAULT").
        """
        self.state_changed()
        return self._state

    def get_value(self):
        """
        Retrieves the current transmission value from the hardware.

        Returns:
            float: The current transmission value, multiplied by 100.
        """
        return self.chan_read_value.get_value() * 100.0

    def state_changed(self, state=None):
        """
        Handles state changes from the hardware.

        Args:
            state (str, optional): The new state from the hardware. If None, the state is fetched from the channel.
        """
        if state is None:
            state = self.chan_state.get_value()

        _str_state = str(state)

        if _str_state == "ON":
            _state = self.STATES.READY
        elif _str_state == "MOVING":
            _state = self.STATES.BUSY
        else:
            _state = self.STATES.FAULT

        self.update_state(_state)

    def _set_value(self, value):
        """
        Sets a new transmission value to the hardware.

        Args:
            value (float): The new transmission value to set.
        """
        value = value / 100.0
        self.chan_set_value.set_value(value)

        while self.get_state() == "MOVING":
            time.sleep(0.1)
            print("Changing transmission")



import logging

from mxcubecore import HardwareRepository as HWR
from mxcubecore.HardwareObjects.MAXIV.MAXIVMD3 import NoPositionBookmarkedError

log = logging.getLogger("user_level_log")


class PrepareOpenHutch:
    """
    Prepare beamline for opening the hutch door.

    - close safety shutter
    - close detector cover
    - move detector to a safe position
    - put MD3 into 'Transfer' phase
    - if jungfrau is used, take pedestal
    """

    def __call__(self):
        try:
            collect = HWR.beamline.collect
            diffractometer = HWR.beamline.diffractometer
            detector = HWR.beamline.detector

            log.info("Preparing experimental hutch for door opening.")

            collect.close_safety_shutter()
            collect.close_detector_cover()

            log.info("Setting diffractometer to transfer phase.")
            diffractometer.wait_device_ready()
            diffractometer.set_phase("Transfer")

            log.info("Moving detector to safe position.")
            collect.move_detector_to_safe_position()

            if detector.get_property("model") == "JUNGFRAU":
                log.info("Collecting Jungfrau pedestal.")
                detector.pedestal()

        except Exception as ex:
            # Explicitly add raised exception into the log message,
            # so that it is shown to the user in the beamline action UI log.
            log.exception("Error preparing to open hutch.\nError was: '%s'", str(ex))  # noqa: TRY401


class MeasureFlux:
    def __call__(self):
        """
        calculate flux at sample position
        """
        flux_at_sample = HWR.beamline.collect.get_instant_flux()
        log.info("Flux at sample position is %.2e ph/s", flux_at_sample)


class SaveMD3Position:
    def __call__(self):
        HWR.beamline.diffractometer.bookmark_position()


class MoveToMD3SavedPosition:
    def __call__(self):
        try:
            HWR.beamline.diffractometer.goto_bookmarked_position()
        except NoPositionBookmarkedError:
            log.warning("No MD3 position saved.")
